import os
from dataclasses import dataclass
from typing import Any, cast

from ..core import Issue, fp, issue, one_of
from ..core.io import CiTool, EnvVars, JsonStr, write_file
from .config import Config, read_config
from .git_env import GitEnv, get_git_env
from .release_env import ReleaseEnv, get_release_env


@dataclass
class MEnv(JsonStr):
    """Contains all the information required for a CI run."""

    config: Config
    env_vars: EnvVars
    git_env: GitEnv
    release_env: ReleaseEnv


def get_m_env(m_dir: str) -> fp.OneOf[Issue, MEnv]:
    """Obtain the M Environment object."""
    return one_of(lambda: [
        MEnv(config, env_vars, git_env, release_env)
        for config in read_config(m_dir)
        for env_vars in CiTool.env_vars()
        for git_env in get_git_env(config, env_vars)
        for release_env in get_release_env(config, env_vars, git_env)
    ]).flat_map_bad(lambda x: issue('get_m_env failure', cause=cast(Issue, x)))


def _m_env_vars(m_env: MEnv) -> fp.OneOf[Issue, str]:
    """Serialize the m environment variables."""
    config = m_env.config
    env_vars = m_env.env_vars
    git = m_env.git_env
    release = m_env.release_env
    associated_pr = (
        git.pull_request or (git.commit and git.commit.associated_pull_request)
    )
    associated_pr_num = associated_pr.pr_number if associated_pr else ''
    env = {
        'M_DIR': config.m_dir,
        'M_OWNER': config.owner,
        'M_REPO': config.repo,
        'M_CI': env_vars.ci_env,
        'M_WORKFLOW': config.workflow.value,
        'M_RUN_ID': env_vars.run_id,
        'M_RUN_NUMBER': env_vars.run_number,
        'M_SHA': git.sha,
        'M_BRANCH': git.branch,
        'M_TARGET_BRANCH': git.target_branch,
        'M_ASSOCIATED_PR_NUMBER': associated_pr_num,
        'M_PR_BRANCH': git.get_pr_branch(),
        'M_PR_NUMBER': git.get_pr_number(),
        'M_TAG': release.build_tag,
        'M_IS_RELEASE': release.is_release,
        'M_IS_RELEASE_PR': release.is_release_pr,
        'M_IS_HOTFIX_PR': release.is_hotfix_pr,
    }
    return fp.Good('\n'.join([f'{key}={val}' for key, val in env.items()]))


def write_m_env_vars(m_dir: str) -> fp.OneOf[Issue, Any]:
    """Write a file with the M environment variables."""
    target_dir = f'{m_dir}/.m'
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    return one_of(lambda: [
        m_env
        for m_env in get_m_env(m_dir)
        for env_list in _m_env_vars(m_env)
        for _ in write_file(f'{m_dir}/.m/env.list', env_list)
    ])



import argparse
import json
import os.path as pth
import sys
from glob import iglob
from typing import Any, Callable, Dict
from typing import MutableMapping as Map
from typing import Optional, Type, Union, cast

from ..core.fp import OneOf
from ..core.io import CiTool, env, error_block
from ..core.issue import Issue
from .validators import validate_non_empty_str


class CmdModule:
    """Interface for command modules."""

    meta: Dict[str, str]

    @staticmethod
    def add_arguments(_parser: argparse.ArgumentParser) -> None:
        """Define options that may apply to the subparsers.

        Should be defined if we want to manipulate the argument parser
        object.
        """

    @staticmethod
    def add_parser(
        _subparser: argparse._SubParsersAction,  # noqa pylint: disable=protected-access
        _raw: Type[argparse.RawTextHelpFormatter],
    ) -> None:
        """Define cli arguments for a command."""

    @staticmethod
    def run(_arg: argparse.Namespace) -> int:
        """Entry point for the cli.

        Call a library function and return 0 if successful or non-zero if there
        is a failure.
        """


def import_mod(name: str) -> CmdModule:
    """Import a module by string."""
    module = __import__(name)
    for part in name.split('.')[1:]:
        module = getattr(module, part)
    return cast(CmdModule, module)


def get_command_modules(
    root: str,
    commands_module: str,
) -> Map[str, CmdModule]:
    """Return a dictionary mapping command names to modules that define an
    `add_parser` method.

    root: The absolute path of the directory containing the __main__.py that
          activates the cli.
    commands_module: The full module resolution. For instance `m.cli.commands`.
    """
    dir_name = '/'.join(commands_module.split('.')[1:])
    mod_names = list(iglob(f'{root}/{dir_name}/*.py'))
    mod = {}
    for name in mod_names:
        tname = pth.split(name)[1][:-3]
        tmod = import_mod(f'{commands_module}.{tname}')
        if hasattr(tmod, 'add_parser'):
            mod[tname] = tmod
    return mod


def get_cli_command_modules(
    file_path: str,
) -> Map[str, Union[CmdModule, Map[str, CmdModule]]]:
    """Return a dictionary containing the commands and subcommands for the cli.

    Note that file_path is expected to be the absolute path to the
    __main__.py file. Another restriction is that the __main__.py file
    must have the `cli.commands` module as its sibling.
    """
    root = pth.split(pth.abspath(file_path))[0]
    main_mod = pth.split(root)[1]
    cli_root = f'{main_mod}.cli'
    root_cmd = get_command_modules(root, f'{cli_root}.commands')
    mod: Map[str, Union[CmdModule, Map[str, CmdModule]]] = {}
    for key, val in root_cmd.items():
        mod[key] = val

    mod['.meta'] = import_mod(f'{cli_root}.commands')
    subcommands = list(iglob(f'{root}/cli/commands/*'))
    for name in subcommands:
        if name.endswith('.py') or name.endswith('__'):
            continue
        tname = pth.split(name)[1]
        mod[tname] = get_command_modules(root, f'{cli_root}.commands.{tname}')
        mod[f'{tname}.meta'] = import_mod(f'{cli_root}.commands.{tname}')
    return mod


def main_parser(
    mod: Map[str, Union[CmdModule, Map[str, CmdModule]]],
    add_args: Optional[Callable[[argparse.ArgumentParser], None]] = None,
):
    """Create an argp and return the result calling its parse_arg method.

    The `add_args` param may be provided as a function that takes in an
    `argparse.ArgumentParser` instance to be able to take additional
    actions.
    """
    meta_mod = cast(CmdModule, mod['.meta'])
    main_meta = meta_mod.meta  # type: ignore
    raw = argparse.RawTextHelpFormatter
    # NOTE: In the future we will need to extend from this class to be able to
    #   override the error method to be able to print CI environment messages.
    argp = argparse.ArgumentParser(
        formatter_class=raw,
        description=main_meta['description'],
    )
    if add_args:
        add_args(argp)
    subp = argp.add_subparsers(
        title='commands',
        dest='command_name',
        required=True,
        help='additional help',
        metavar='<command>',
    )
    names = sorted(mod.keys())
    for name in names:
        if name.endswith('.meta'):
            continue
        if isinstance(mod[name], dict):
            meta_mod = cast(CmdModule, mod[f'{name}.meta'])
            meta = meta_mod.meta  # type: ignore
            parser = subp.add_parser(
                name,
                help=meta['help'],
                formatter_class=raw,
                description=meta['description'],
            )
            if hasattr(meta_mod, 'add_arguments'):
                meta_mod.add_arguments(parser)
            subsubp = parser.add_subparsers(
                title='commands',
                dest='subcommand_name',
                required=True,
                help='additional help',
                metavar='<command>',
            )
            sub_mod = cast(Dict[str, CmdModule], mod[name])
            for subname in sorted(sub_mod.keys()):
                sub_mod[subname].add_parser(subsubp, raw)
        else:
            cast(CmdModule, mod[name]).add_parser(subp, raw)
    return argp.parse_args()


def run_cli(
    file_path: str,
    main_args=None,
) -> None:
    """Helper function to create a cli application.

        def main_args(argp):
            argp.add_argument(...)
        def main():
            run_cli(__file__, main_args)

    We only need `main_args` if we need to gain access to the
    `argparse.ArgumentParser` instance.
    """
    mod = get_cli_command_modules(file_path)
    arg = main_parser(mod, main_args)
    if hasattr(arg, 'subcommand_name'):
        sub_mod = cast(Dict[str, CmdModule], mod[arg.command_name])
        sys.exit(sub_mod[arg.subcommand_name].run(arg))
    else:
        sys.exit(cast(CmdModule, mod[arg.command_name]).run(arg))


def display_issue(issue: Issue) -> None:
    """print an error message."""
    CiTool.error(issue.message)
    error_block(str(issue))


def display_result(val: Any) -> None:
    """print the JSON stringification of the param `val` provided that val is
    not `None`."""
    if val is not None:
        try:
            print(json.dumps(val, separators=(',', ':')))
        except Exception:
            print(val)


def run_main(
    callback: Callable[[], OneOf[Issue, Any]],
    handle_result: Callable[[Any], None] = display_result,
    handle_issue: Callable[[Issue], None] = display_issue,
):
    """Run the callback and print the returned value as a JSON string. Set the
    print_raw param to True to bypass the JSON stringnification. To change how
    the result or an issue should be display then provide the optional
    arguments handle_result and handle_issue. For instance, to display the raw
    value simply provide the `print` function.

    Return 0 if the callback is a `Good` result otherwise return 1.
    """
    try:
        res = callback()
        val = res.value
        if res.is_bad:
            if isinstance(val, Issue):
                handle_issue(val)
            else:
                issue = Issue('non-issue exception', cause=cast(Issue, val))
                handle_issue(issue)
            return 1
        handle_result(val)
    except Exception as ex:
        issue = Issue('unknown caught exception', cause=ex)
        handle_issue(issue)
        return 1
    return 0


def error(msg: str, issue: Optional[Issue] = None) -> int:
    """print an error message."""
    CiTool.error(msg)
    if issue:
        error_block(str(issue))
    return 1


def cli_integration_token(integration: str, env_var: str):
    """Return a function that takes in a parser.

    This generated function registers a token argument in the parser
    which looks for its value in the environment variables.
    """
    return lambda parser: parser.add_argument(
        '-t',
        '--token',
        type=validate_non_empty_str,
        default=env(env_var),
        help=f'{integration} access token (default: env.{env_var})',
    )



import inspect

from ...utils import env, run_main


def add_parser(sub_parser, raw):
    desc = """
        Merge a pull request

        https://docs.github.com/en/rest/reference/pulls#merge-a-pull-request

        example:

            $ m github merge_pr \\
                --owner owner \\
                --repo repo \\
                --pr pr_number \\
                'commit_title' | m json
    """
    parser = sub_parser.add_parser(
        'merge_pr',
        help='merge a pull request',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--owner',
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner (default: env.GITHUB_REPOSITORY_OWNER)',
    )
    add(
        '--repo',
        required=True,
        help='repo name',
    )
    add(
        '--commit-title',
        help='commit title',
    )
    add(
        'pr',
        type=int,
        help='the pr number',
    )


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....github.api import merge_pr
    return run_main(lambda: merge_pr(
        arg.token,
        arg.owner,
        arg.repo,
        arg.pr,
        arg.commit_title,
    ))



import inspect

from ...utils import env, run_main


def add_parser(sub_parser, raw):
    desc = """
        Retrieve the information required for continuous integration.

        example:

            $ m github ci \\
                --owner jmlopez-rod \\
                --repo pysync \\
                --sha 4538b2a2556efcbdfc1e7df80c4f71ade45f3958 \\
                --pr 1 \\
                --include-release \\
            | m json
            {
            "commit": {
                "associatedPullRequests": {
                "nodes": [
                    {
                    "author": {
                        "login": "jmlopez-rod",
                        "avatarUrl": "https://avatars.githubusercontent.com/u/1810397?s=50&v=4",
                        "email": ""
                    },
            ...

        NOTE: Use the --merge-commit flag if you are providing a sha from
        github actions.
    """  # noqa
    parser = sub_parser.add_parser(
        'ci',
        help='continous integration information',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--owner',
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner (default: env.GITHUB_REPOSITORY_OWNER)',
    )
    add(
        '--repo',
        required=True,
        help='repo name',
    )
    add(
        '--sha',
        required=True,
        help='commit sha',
    )
    add(
        '--pr',
        type=int,
        help='pull request number',
    )
    add(
        '--file-count',
        default=10,
        type=int,
        help='max number of files to retrieve',
    )
    add(
        '--include-release',
        default=False,
        action='store_true',
        help='include the last release information',
    )
    add(
        '--merge-commit',
        default=False,
        action='store_true',
        help='set if the sha is a merge commit sha (from github)',
    )


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....github.ci import CommitInfo, get_raw_ci_run_info
    return run_main(lambda: get_raw_ci_run_info(
        arg.token,
        CommitInfo(arg.owner, arg.repo, arg.sha),
        arg.pr,
        arg.file_count,
        arg.include_release,
        arg.merge_commit,
    ))



from .core import one_of, subprocess
from .core.fp import OneOf
from .core.issue import Issue


def get_branch() -> OneOf[Issue, str]:
    """Get the current git branch name.

    Returns:
        A `OneOf` containing an `Issue` or a string specifying the branch.
    """
    return subprocess.eval_cmd('git rev-parse --abbrev-ref HEAD')


def get_first_commit_sha() -> OneOf[Issue, str]:
    """Find the first commit sha in the repository.

    Returns:
        A `OneOf` containing an `Issue` or a string of the first commit sha.
    """
    return subprocess.eval_cmd('git rev-list --max-parents=0 HEAD')


def get_current_commit_sha() -> OneOf[Issue, str]:
    """Find the sha of the current commit.

    Returns:
        A `OneOf` containing an `Issue` or a string of the current commit sha.
    """
    return subprocess.eval_cmd('git rev-parse HEAD')


def get_remote_url() -> OneOf[Issue, str]:
    """Find the remote url of the repo.

    Returns:
        A `OneOf` containing an `Issue` or a string with the url.
    """
    return subprocess.eval_cmd('git config --get remote.origin.url')


def _extract_status(msg: str) -> str:
    matches = [
        ('Untracked files', 'untracked'),
        ('Your branch is ahead', 'ahead'),
        ('Your branch is behind', 'behind'),
        ('working tree clean', 'clean'),
        ('Changes to be committed', 'staged'),
        ('Changed but not updated', 'dirty'),
        ('Changes not staged', 'dirty'),
        ('Unmerged paths', 'dirty'),
        ('diverged', 'diverged'),
    ]
    for substr, key in matches:
        if substr in msg:
            return key
    return '?'


def get_status() -> OneOf[Issue, str]:
    """Find the current git status.

    Returns:
        A `OneOf` containing an `Issue` or a word denoting the git status.
    """
    res = subprocess.eval_cmd('git status')
    return one_of(
        lambda: [
            _extract_status(msg)
            for msg in res
        ],
    )



from typing import Dict

from ...core import Good, Issue, OneOf, issue
from .core.process import PostProcessor
from .core.types import Configuration, Transform
from .post_processors import eslint, pycodestyle, pylint


def get_post_processor(
    name: str,
    celt_config: Configuration,
) -> OneOf[Issue, PostProcessor]:
    """Find an available post processor based on the key provided.

    Args:
        name: name of the post processor
        celt_config: The configuration to use

    Returns:
        A `OneOf` containing an `Issue` or a post processor function.
    """
    mapping: Dict[str, Transform] = {
        'eslint': eslint.read_payload,
        'pycodestyle': pycodestyle.read_payload,
        'flake8': pycodestyle.read_payload,
        'pylint': pylint.read_payload,
    }
    if name not in mapping:
        return issue(f'{name} is not a supported post processor')

    return Good(PostProcessor(name, celt_config, mapping[name]))



from typing import Any, Callable, Dict, List, Optional, TypeVar, cast

from .fp import Bad, Good, OneOf, StopBadIteration
from .issue import Issue

G = TypeVar('G')  # pylint: disable=invalid-name


def issue(  # noqa: WPS211
    message: str,
    description: Optional[str] = None,
    cause: Optional[Exception] = None,
    context: Optional[object] = None,
    data: Optional[object] = None,  # noqa: WPS110
    include_traceback: bool = True,
) -> OneOf[Issue, Any]:
    """Shortcut to create a Bad OneOf containing an Issue.

    Args:
        message: The issue message.
        description: Optional description.
        cause: Optional exception that caused the issue.
        context: Optional dictionary to provide extra information.
        data: deprecated, use context instead.
        include_traceback: Defaults to true to provde the stack trace.

    Returns:
        An instance of an `Issue`.
    """
    return Bad(
        Issue(message, description, cause, context or data, include_traceback),
    )


def one_of(comp: Callable[[], List[G]]) -> OneOf[Any, G]:
    """Handle the "Good" value of a `OneOf`.

    To be used so that we may iterate over OneOf objects that may raise
    the `StopBadIteration` exception.

    Args:
        comp: A lambda function returning an array with a single value.

    Returns:
        A `OneOf` with the value returned from `comp`.
    """
    try:
        return Good(comp()[0])
    except StopBadIteration as ex:
        return cast(Bad, ex.bad)
    except Exception as ex:
        return issue('one_of caught exception', cause=ex)


def to_one_of(
    callback: Callable[[], Any],
    message: str,
    context: Optional[Dict[str, Any]] = None,
) -> OneOf[Issue, int]:
    """Wrap a python call in a `OneOf`.

    Args:
        callback: A lambda function with a simple python statement.py
        message: An error message in case the statment raises an exception.
        context: Additional error context.

    Returns:
        A `OneOf` containing an `Issue` if the callback raises an error.
    """
    try:
        callback()
    except Exception as ex:
        return issue(message, cause=ex, context=context)
    return Good(0)



from inspect import cleandoc

from ...utils import cli_integration_token

_DESC = """
    The following commands make calls to the Github api.
"""

meta = {
    'help': 'call the github api',
    'description': cleandoc(_DESC),
}

add_arguments = cli_integration_token('github', 'GITHUB_TOKEN')



import inspect
import json
import sys
import traceback
from collections import OrderedDict
from typing import List, Optional, cast

from typing_extensions import TypedDict

IssueDict = TypedDict(
    'IssueDict',
    {
        'message': str,
        'description': str,
        'cause': object,
        'data': object,
        'traceback': List[str],
    },
    total=False,
)


def remove_traceback(obj: object) -> None:
    """Remove the `traceback` key from a dictionary if it exists and
    recursively remove the `traceback` from its cause."""
    if isinstance(obj, dict):
        if 'traceback' in obj:
            del obj['traceback']
        if 'cause' in obj:
            remove_traceback(obj['cause'])


class Issue(Exception):  # noqa: N818 - Intention is not to raise
    """Wrapper to keep track of all exceptions.

    It provides a 'cause' field so that we may know why an issue was
    triggered.
    """

    show_traceback = True

    message: str
    description: Optional[str]
    cause: Optional[Exception]
    data: Optional[object]
    include_traceback: bool
    cause_tb: Optional[List[str]]

    def __init__(
        self,
        message: str,
        description: Optional[str] = None,
        cause: Optional[Exception] = None,
        data: Optional[object] = None,
        include_traceback: bool = True,
    ):
        """Create an Issue.

        - description: More in depth detail on the issue
        - cause: The exception that or Issue that is responsible for this
                 Issue instance.
        - data: Provide a dictionary with any useful data related to the issue.
        - include_traceback: If False, it skips computing the traceback.
        """
        super().__init__()
        self.message = message
        self.description = description
        self.cause = cause
        if cause and not isinstance(cause, Issue):
            # https://stackoverflow.com/a/12539332/788553
            try:
                exception_list = [
                    *traceback.format_tb(sys.exc_info()[2]),
                    *traceback.format_exception_only(
                        sys.exc_info()[0],
                        sys.exc_info()[1],
                    ),
                ]
                self.cause_tb = [
                    y
                    for x in exception_list
                    for y in x.splitlines()
                ]
            finally:
                pass
        self.data = data
        self.include_traceback = include_traceback
        if self.include_traceback:
            frame = inspect.currentframe()
            try:
                self.traceback = [
                    y
                    for x in traceback.format_stack(frame)[:-1]
                    for y in x.splitlines()
                ]
            finally:
                del frame

    def to_dict(self) -> IssueDict:
        """Convert to a ordered dictionary so that each of the properties are
        written in an expected order.
        """
        obj = cast(IssueDict, OrderedDict())
        obj['message'] = self.message
        if self.description:
            obj['description'] = self.description
        if self.data:
            obj['data'] = self.data
        if self.include_traceback:
            obj['traceback'] = self.traceback
        if self.cause:
            if isinstance(self.cause, Issue):
                obj['cause'] = self.cause.to_dict()
            else:
                obj['cause'] = {
                    'message': str(self.cause),
                    'traceback': self.cause_tb,
                }
        return obj

    def to_str(self, show_traceback: bool) -> str:
        """Convert the instance to string.

        We have the option of not showing the traceback.
        """
        obj = self.to_dict()
        if not show_traceback:
            remove_traceback(obj)
        return json.dumps(obj, indent=2)

    def __str__(self) -> str:
        return self.to_str(Issue.show_traceback)



import inspect


def add_parser(sub_parser, raw):
    desc = 'close and immediately open another block'
    parser = sub_parser.add_parser(
        'sibling_block',
        help='close and open a sibling block',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add('to_close', help='block name to close')
    add('name', help='block name to open')
    add('description', help='new block description')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....core.io import CiTool
    CiTool.close_block(arg.to_close)
    CiTool.open_block(arg.name, arg.description)
    return 0



import inspect


def add_parser(sub_parser, raw):
    desc = """
        close a block. When a block is closed, all its inner blocks are
        closed automatically. Note: Not all CI Tools support nesting.
    """
    parser = sub_parser.add_parser(
        'close',
        help='close block',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    parser.add_argument('name', type=str, help='block name to close')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....core.io import CiTool
    CiTool.close_block(arg.name)
    return 0



import inspect

from ..validators import validate_json_payload


def add_parser(sub_parser, raw):
    desc = r"""
        query json data.

        single value:

            m jsonq path.to.property < file.json
            cat file.json | m jsonq path.to.property
            m jsonq @file.json path.to.property

        Return the value stored in the json file. For arrays and objects it
        will print the python representation of the object.

        multiple values:

            m jsonq path1 path2 path3 < file.json
            cat file.json | m jsonq path1 path2 path3
            m jsonq @file.json path1 path2 path3

        use `read` to store in bash variables:

            read -r -d '\n' \
                var1 var2 var3 \
                <<< "$(m jsonq @file.json 'path1' 'path2' 'path3')"
    """
    parser = sub_parser.add_parser(
        'jsonq',
        help='query json data',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        'payload',
        type=validate_json_payload,
        nargs='?',
        default='@-',
        help='json data: @- (default stdin), @filename (file), string',
    )
    add(
        '-w',
        '--warn',
        action='store_true',
        help='print warning messages instead of errors',
    )
    add(
        '-s',
        '--separator',
        type=str,
        default='\n',
        help=r'separator for multiple values (default \n)',
    )
    add('query', type=str, nargs='+', help='path to json data')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ...core.json import jsonq
    return jsonq(arg.payload, arg.separator, arg.warn, *arg.query)



def commit_query(include_pr: bool, include_author: bool) -> str:
    """Build a graphql query for github.

    The output of this function is meant to go inside the repository
    field.
    """
    author = """
      author {
        name
        email
        user {
          login
        }
      }
      oid
    """ if include_author else ''
    pr = """
      associatedPullRequests(first: 1) {
        nodes {
          author {
            login
            avatarUrl(size: 50)
            ... on User {
              email
            }
          }
          number
          title
          body
          baseRefName
          baseRefOid
          headRefName
          headRefOid
          merged
        }
      }
    """ if include_pr else ''
    query = """
      commit: object(expression: $sha) {
        ... on Commit {
          message
          %s
          %s
        }
      }
    """
    return query % (author, pr)


LATEST_RELEASE = """
  releases(last: 1) {
    nodes {
      name
      tagName
      publishedAt
    }
  }
"""

PULL_REQUEST = """
  pullRequest(number: $pr) {
    headRefName
    headRefOid
    baseRefName
    baseRefOid
    title
    body
    url
    author {
      login
      avatarUrl(size: 50)
      ... on User {
        email
      }
    }
    files(first: $fc) {
      totalCount
      nodes {
        path
      }
    }
    merged
    isDraft
  }
"""



from dataclasses import dataclass

from ..core import issue, one_of
from ..core.fp import Good, OneOf
from ..core.io import EnvVars, JsonStr
from ..core.issue import Issue
from .config import Config, Workflow
from .git_env import GitEnv


@dataclass
class ReleaseEnv(JsonStr):
    """Object to store the release configuration."""

    build_tag: str
    is_release: bool
    is_release_pr: bool
    is_hotfix_pr: bool
    workflow: Workflow


def _verify_version(
    config: Config,
    git_env: GitEnv,
    gh_latest: str,
    is_release_pr: bool,
    is_hotfix_pr: bool,
    is_release: bool,
) -> OneOf[Issue, int]:
    if config.workflow in [Workflow.git_flow, Workflow.m_flow]:
        if config.uses_git_flow():
            pr_branch = git_env.get_pr_branch()
            flow = config.git_flow
            # Skip verification when release or hotfix are going to develop
            if git_env.target_branch == flow.develop_branch:
                if (
                    pr_branch.startswith(flow.release_prefix) or
                    pr_branch.startswith(flow.hotfix_prefix)
                ):
                    return Good(0)
        return config.verify_version(
            gh_latest,
            is_release_pr=(is_release_pr or is_hotfix_pr),
            is_release=is_release,
        )
    # Covers Workflow.free_flow
    return Good(0)


def _get_master_branch(config: Config) -> str:
    if config.uses_git_flow():
        return config.git_flow.master_branch
    if config.uses_m_flow():
        return config.m_flow.master_branch
    return 'master'


def _get_develop_branch(config: Config) -> str:
    if config.uses_git_flow():
        return config.git_flow.develop_branch
    return 'develop'


def get_release_env(
    config: Config,
    env_vars: EnvVars,
    git_env: GitEnv,
) -> OneOf[Issue, ReleaseEnv]:
    """Provide the release environment information."""
    is_release = git_env.is_release(config)
    is_release_pr = git_env.is_release_pr(config)
    is_hotfix_pr = git_env.is_hotfix_pr(config)
    gh_latest = git_env.release.tag_name if git_env.release else ''
    if not config.uses_free_flow():
        master_branch = _get_master_branch(config)
        develop_branch = _get_develop_branch(config)
        if config.uses_git_flow():
            if (
                (is_release_pr or is_hotfix_pr) and
                git_env.target_branch not in (master_branch, develop_branch)
            ):
                return issue(f'hotfix and releases prs to {master_branch}')
            if (
                is_release and
                git_env.branch not in (master_branch, develop_branch)
            ):
                return issue(f'hotfix and releases only on {master_branch}')
        if config.uses_m_flow():
            if (
                is_release_pr and
                git_env.target_branch != master_branch
            ):
                return issue(f'release prs to {master_branch}')
            if (
                is_release and
                git_env.branch != master_branch
            ):
                return issue(f'releases only on {master_branch}')
    return one_of(lambda: [
        ReleaseEnv(
            build_tag=build_tag,
            is_release=is_release,
            is_release_pr=is_release_pr,
            is_hotfix_pr=is_hotfix_pr,
            workflow=config.workflow,
        )
        for _ in _verify_version(
            config,
            git_env,
            gh_latest,
            is_release_pr=is_release_pr,
            is_hotfix_pr=is_hotfix_pr,
            is_release=is_release)
        for build_tag in git_env.get_build_tag(config, env_vars.run_id)
    ])



from dataclasses import dataclass
from typing import Any, Mapping, Optional

from ..core import issue, one_of
from ..core.fp import Good, OneOf
from ..core.http import fetch_json
from ..core.issue import Issue


def request(
    token: str,
    endpoint: str,
    method: str = 'GET',
    data: Optional[Any] = None,
) -> OneOf[Issue, Any]:
    """Make an api request to github. See:

    - https://docs.github.com/en/rest/overview/resources-in-the-rest-api
    - https://docs.github.com/en/rest/overview/endpoints-available-for-github-apps
    """  # noqa
    url = f'https://api.github.com{endpoint}'
    headers = {'authorization': f'Bearer {token}'}
    return fetch_json(url, headers, method, data)


def _filter_data(data: Mapping[str, Any]) -> OneOf[Issue, Any]:
    if data.get('data'):
        return Good(data['data'])
    return issue(
        'github response missing data field',
        context={'response': data},
    )


def graphql(
    token: str,
    query: str,
    variables: Mapping[str, Any],
) -> OneOf[Issue, Any]:
    """Make a request to Github's graphql API:

    https://docs.github.com/en/graphql/guides/forming-calls-with-graphql
    """
    data = {'query': query, 'variables': variables or {}}
    return one_of(
        lambda: [
            payload
            for res in request(token, '/graphql', 'POST', data)
            for payload in _filter_data(res)
        ],
    )


def create_release(
    token: str,
    owner: str,
    repo: str,
    version: str,
    branch: Optional[str] = None,
) -> OneOf[Issue, Any]:
    """Send a payload to create a release in github."""
    endpoint = f'/repos/{owner}/{repo}/releases'
    base = 'https://github.com'
    link = f'{base}/{owner}/{repo}/blob/master/CHANGELOG.md#{version}'
    data = {
        'tag_name': version,
        'name': f'{version}',
        'body': f'**See [CHANGELOG]({link}).**',
        'draft': False,
        'prerelease': False,
    }
    if branch:
        data['target_commitish'] = branch
    return request(token, endpoint, 'POST', data)


@dataclass
class GithubPullRequest:
    """Data needed to create a pull request."""

    title: str
    body: str
    head: str
    base: str


def create_pr(
    token: str,
    owner: str,
    repo: str,
    pr_info: GithubPullRequest,
) -> OneOf[Issue, Any]:
    """Send a payload to create a pull request in github."""
    endpoint = f'/repos/{owner}/{repo}/pulls'
    data = {
        'title': pr_info.title,
        'body': pr_info.body,
        'head': pr_info.head,
        'base': pr_info.base,
    }
    return request(token, endpoint, 'POST', data)


def merge_pr(
    token: str,
    owner: str,
    repo: str,
    pr_number: int,
    commit_title: Optional[str],
) -> OneOf[Issue, Any]:
    """Send a payload to merge a pull request in github."""
    endpoint = f'/repos/{owner}/{repo}/pulls/{pr_number}/merge'
    data = {'commit_title': commit_title} if commit_title else {}
    return request(token, endpoint, 'PUT', data)


@dataclass
class GithubShaStatus:
    """Data needed to create a pull request."""

    sha: str
    context: str
    state: str
    description: str
    url: Optional[str] = None


def commit_status(
    token: str,
    owner: str,
    repo: str,
    sha_info: GithubShaStatus,
):
    """Set a status for a sha. The valid states are:

    - pending
    - success
    - failure
    - error
    """
    endpoint = f'/repos/{owner}/{repo}/statuses/{sha_info.sha}'
    data = {
        'context': sha_info.context,
        'state': sha_info.state,
        'description': sha_info.description,
    }
    if sha_info.url:
        data['target_url'] = sha_info.url
    return request(token, endpoint, 'POST', data)



from inspect import cleandoc as cdoc
from typing import Tuple

from ...argparse import Arg, add_arguments
from ...utils import run_main

DESC = """
    Remove empty npm tags.

        $ m npm clean_tags @scope/package

    When packages are removed there will be empty tags that point to nothing.
    This command will find those empty tags and remove them.
"""
ARGS: Tuple[Arg, ...] = (
    (
        ['package_name'],
        'name of the npm package',
        {'type': str},
    ),
)


def add_parser(sub_parser, raw):
    parser = sub_parser.add_parser(
        'clean_tags',
        help='remove empty npm tags',
        formatter_class=raw,
        description=cdoc(DESC),
    )
    add_arguments(parser, ARGS)


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....npm.clean_tags import clean_tags
    return run_main(lambda: clean_tags(arg.package_name))



from inspect import cleandoc

_DESC = """
    The following commands provide local support in the continuous
    integration flow.
"""

meta = {
    'help': 'continuous integration utilities',
    'description': cleandoc(_DESC),
}



from typing import Any

from ..core import one_of
from ..core.fp import Good, OneOf
from ..core.issue import Issue
from ..core.json import get
from .api import graphql
from .ci import create_ci_query


def get_pr_info(
    token: str,
    owner: str,
    repo: str,
    pr_number: int,
    file_count: int,
) -> OneOf[Issue, Any]:
    """Retrieve the information of the given Github PR."""
    query = create_ci_query(pr_number, False, False)
    variables = {
        'owner': owner,
        'repo': repo,
        'pr': pr_number,
        'fc': file_count,
    }
    return one_of(
        lambda: [
            data
            for res in graphql(token, query, variables)
            for data in get(res, 'repository.pullRequest')
        ],
    )


def get_latest_release(
    token: str,
    owner: str,
    repo: str,
) -> OneOf[Issue, str]:
    """Retrieve the latest release for a repo."""
    query = """query ($owner: String!, $repo: String!) {
      repository(owner:$owner, name:$repo) {
         releases(last: 1) {
            nodes {
                name
                tagName
                publishedAt
            }
        }
      }
    }"""
    variables = {'owner': owner, 'repo': repo}
    return one_of(
        lambda: [
            data
            for res in graphql(token, query, variables)
            for data in get(res, 'repository.releases.nodes.0.tagName')
        ],
    ).flat_map_bad(lambda _: Good('0.0.0'))



from ...utils import run_main


def add_parser(sub_parser, raw):
    parser = sub_parser.add_parser(
        'env',
        help='create a list of env variables',
        formatter_class=raw,
        description='Create the [m_dir]/.m/env.list file',
    )
    add = parser.add_argument
    add('m_dir', type=str, help='m project directory')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....ci.m_env import write_m_env_vars
    return run_main(lambda: write_m_env_vars(arg.m_dir))



def compare_sha_url(owner: str, repo: str, prev: str, next_: str) -> str:
    """Provide a url to compare two sha/tags in a github repo."""
    return f'https://github.com/{owner}/{repo}/compare/{prev}...{next_}'



import re
from datetime import datetime
from typing import List

from ..ci.config import Config, read_config
from ..core import issue, one_of
from ..core.fp import Good, OneOf
from ..core.io import read_file, write_file
from ..core.issue import Issue
from ..git import get_first_commit_sha
from ..github import compare_sha_url


def _get_versions(lines: List[str], new_ver: str, first_sha: str) -> List[str]:
    versions = [new_ver]
    for line in lines:
        match = re.match(r'## \[(.*)]', line)
        if match:
            versions.append(match.groups()[0])
    versions.append(first_sha)
    return versions


def _version_anchor(ver: str) -> str:
    return f'<a name="{ver}" href="#{ver}">-</a>'


def new_changelog(
    content: str,
    owner: str,
    repo: str,
    new_ver: str,
    first_sha: str,
) -> OneOf[Issue, str]:
    """Modify the contents of a CHANGELOG so that a new entry with the new
    version is added to the new changelog contents."""
    parts = content.split('## [Unreleased]')
    if len(parts) != 2:
        return issue('missing "Unreleased" link')

    header, main = parts
    entries = main.split('[unreleased]:')[0]
    versions = _get_versions(entries.split('\n'), new_ver, first_sha)

    links = [f'[unreleased]: {compare_sha_url(owner, repo, new_ver, "HEAD")}']
    for i in range(len(versions) - 1):
        link = compare_sha_url(owner, repo, versions[i + 1], versions[i])
        links.append(f'[{versions[i]}]: {link}')

    date = datetime.now().strftime('%B %d, %Y')
    return Good(''.join([
        header,
        '## [Unreleased]\n\n',
        f'## [{new_ver}] {_version_anchor(new_ver)} {date}\n\n',
        entries,
        '\n'.join(links),
        '\n',
    ]))


def update_changelog_file(
    owner: str,
    repo: str,
    new_ver: str,
    first_sha: str,
    filename: str = 'CHANGELOG.md',
) -> OneOf[Issue, int]:
    """Add an entry to the CHANGELOG file with the new version to be
    released."""
    return one_of(lambda: [
        0
        for data in read_file(filename)
        for new_data in new_changelog(data, owner, repo, new_ver, first_sha)
        for _ in write_file(filename, new_data)
    ])


def _update_line_version(line: str, ver: str) -> str:
    content = line.strip()
    if not content.startswith('"version"'):
        return line
    comma = ',' if content.endswith(',') else ''
    return f'  "version": "{ver}"{comma}'


def _update_config_version(contents: str, ver: str) -> OneOf[Issue, str]:
    lines = contents.split('\n')
    new_lines = [_update_line_version(line, ver) for line in lines]
    return Good('\n'.join(new_lines))


def update_version(root: str, version: str) -> OneOf[Issue, int]:
    """Update the version property in m.json configuration file."""
    filename = f'{root}/m.json'
    return one_of(lambda: [
        0
        for data in read_file(filename)
        for new_data in _update_config_version(data, version)
        for _ in write_file(filename, new_data)
    ])


def _success_release_setup(config: Config, new_ver: str) -> OneOf[Issue, int]:
    link = compare_sha_url(config.owner, config.repo, config.version, 'HEAD')
    print(f'\nSetup for version {new_ver} is complete.')
    print(f'Unreleased changes: {link}\n')
    return Good(0)


def release_setup(
    m_dir: str,
    new_ver: str,
    changelog: str = 'CHANGELOG.md',
) -> OneOf[Issue, None]:
    """Modify all the necessary files to create a release."""
    return one_of(lambda: [
        None
        for config in read_config(m_dir)
        for first_sha in get_first_commit_sha()
        for _ in update_version(m_dir, new_ver)
        for _ in update_changelog_file(
            config.owner,
            config.repo,
            new_ver,
            first_sha,
            changelog)
        for _ in _success_release_setup(config, new_ver)
    ])



import sys
from argparse import ArgumentTypeError
from pathlib import Path

from ..core.json import parse_json, read_json


def validate_json_payload(file_path: str):
    """Return a dictionary from the contents of file_path.

    This is a string that tell us to read from a file, stdin or just
    plain json data.

    Args:
        file_path: A string with the json payload. If it starts with `@` then
            the name of a valid json file from where the payload will be read.

    Raises:
        ArgumentTypeError: If the file_path is meant to be a valid path and it
            does not exist.

    Returns:
        A parsed json payload
    """
    if file_path == '@-':
        res = read_json(None)
        if res.is_bad:
            msg = f'invalid json payload in SYS.STDIN\n{res.value}'
            raise ArgumentTypeError(msg)
        return res.value
    if file_path.startswith('@'):
        filename = file_path[1:]
        if not Path(filename).exists():
            raise ArgumentTypeError(f'file "{filename}" does not exist')
        res = read_json(filename)
        if res.is_bad:
            raise ArgumentTypeError(f'invalid json payload in {filename}')
        return res.value
    res = parse_json(file_path)
    if res.is_bad:
        raise ArgumentTypeError(f'invalid json payload\n{res.value}')
    return res.value


def validate_payload(file_path: str) -> str:
    """Return the raw payload.

    This allows us to read from a file or the stdin stream.

    Args:
        file_path: A string with the payload. If it starts with `@` then the
            name of a valid file from where the payload will be read.

    Raises:
        ArgumentTypeError: If the file_path is meant to be a valid path and it
            does not exist.

    Returns:
        The payload found in the file.
    """
    if file_path.startswith(r'\@'):
        # escape @ with \ to let the cli know that the payload starts with @
        return file_path[1:]
    if file_path == '@-':
        return sys.stdin.read()
    if file_path.startswith('@'):
        filename = file_path[1:]
        if not Path(filename).exists():
            raise ArgumentTypeError(f'file "{filename}" does not exist')
        with open(filename, encoding='UTF-8') as fp:
            return fp.read()
    return file_path


def validate_non_empty_str(arg_value: str) -> str:
    """Return the value as long as its not empty.

    Args:
        arg_value: The input provided by argparse.

    Raises:
        ArgumentTypeError: If the input is not provided.

    Returns:
        The value if non empty.
    """
    if not arg_value:
        raise ArgumentTypeError('empty value not allowed')
    return arg_value



from dataclasses import dataclass
from typing import List, Optional

from ..core.io import JsonStr


@dataclass
class Author(JsonStr):
    """An object representing a commiter."""

    login: str
    avatar_url: str
    email: str


@dataclass
class AssociatedPullRequest(JsonStr):
    """Information for commits that are associated with a pull request."""

    # pylint: disable=too-many-instance-attributes
    author: Author
    merged: bool
    pr_number: int
    target_branch: str
    target_sha: str
    pr_branch: str
    pr_sha: str
    title: str
    body: str


@dataclass
class Commit(JsonStr):
    """The git commit info."""

    author_login: str
    short_sha: str
    sha: str
    message: str
    url: str
    associated_pull_request: Optional[AssociatedPullRequest] = None

    def get_pr_branch(self) -> str:
        """Return the pr branch if the commit has an associated pr or empty
        string."""
        if not self.associated_pull_request:
            return ''
        return self.associated_pull_request.pr_branch

    def is_release(self, release_prefix: Optional[str]) -> bool:
        """Determine if the current commit should create a release."""
        if not release_prefix:
            return False
        return self.get_pr_branch().startswith(release_prefix)


@dataclass
class PullRequest(JsonStr):
    """Pull request information."""

    # pylint: disable=too-many-instance-attributes
    author: Author
    pr_number: int
    pr_branch: str
    target_branch: str
    target_sha: str
    url: str
    title: str
    body: str
    file_count: int
    files: List[str]
    is_draft: bool

    def is_release_pr(self, release_prefix: Optional[str]) -> bool:
        """Determine if the pull request is a release pull request."""
        if not release_prefix:
            return False
        return self.pr_branch.startswith(release_prefix)


@dataclass
class Release(JsonStr):
    """Github release <==> Git tag."""

    name: str
    tag_name: str
    published_at: str


@dataclass
class GithubCiRunInfo(JsonStr):
    """The main information we need for a ci run."""

    commit: Commit
    pull_request: Optional[PullRequest] = None
    release: Optional[Release] = None


@dataclass
class CommitInfo(JsonStr):
    """A commit can be tracked with the following properties."""

    owner: str
    repo: str
    sha: str



import inspect

from ...utils import run_main


def add_parser(sub_parser, raw):
    desc = """
        Display the current git branch name.

            $ m git branch
            master
    """
    sub_parser.add_parser(
        'branch',
        help='display the current git branch',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )


def run(_):
    # pylint: disable=import-outside-toplevel
    from .... import git
    return run_main(git.get_branch, print)



import os
import re
from inspect import cleandoc as cdoc
from pathlib import Path
from typing import List, Tuple

from ..core import Good, Issue, OneOf, issue, one_of
from ..core.io import CiTool, read_file, write_file
from ..core.subprocess import eval_cmd
from ..git import get_remote_url


def parse_ssh_url(ssh_url: str) -> OneOf[Issue, Tuple[str, str]]:
    """Find the owner and repo from an ssh url.

    Args:
        ssh_url: The url of the repo

    Returns:
        A tuple with the owner and repo (or an Issue).
    """
    match = re.findall('.*:(.*)/(.*).git', ssh_url)
    if match:
        return Good(match[0])
    return issue(
        'unable to obtain owner and repo',
        context={'ssh_url': ssh_url},
    )


def get_repo_info() -> OneOf[Issue, Tuple[str, str]]:
    """Get the owner and repo name from the current repository.

    Returns:
        A tuple with the owner and repo (or an Issue).
    """
    return one_of(lambda: [
        owner_repo
        for ssh_url in get_remote_url()
        for owner_repo in parse_ssh_url(ssh_url)
    ])


def m_json_body(owner: str, repo: str) -> str:
    """Create the basic contents of a m configuration file.

    Args:
        owner: The repo owner.
        repo: The repo name.

    Returns:
        A json string to be the content of the `m.json` file.
    """
    body = f"""
        {{
          "owner": "{owner}",
          "repo": "{repo}",
          "version": "0.0.0"
        }}
    """
    cbody = cdoc(body)
    return f'{cbody}\n'


def create_m_config() -> OneOf[Issue, int]:
    """Create the m configuration file.

    Returns:
        A `OneOf` containing 0 if successful or an `Issue`.
    """
    if not os.path.exists('m'):
        os.makedirs('m')
    return one_of(lambda: [
        0
        for owner, repo in get_repo_info()
        for _ in write_file('m/m.json', m_json_body(owner, repo))
    ])


def create_changelog() -> OneOf[Issue, int]:
    """Create the changelog file.

    Returns:
        A `OneOf` containing 0 if successful or an `Issue`.
    """
    body = """
        # Changelog

        The format of this changelog is based on [Keep a Changelog](http://keepachangelog.com/en/1.0.0/).
        The project adheres to [Semantic Versioning](http://semver.org/spec/v2.0.0.html)

        > Major version zero (0.y.z) is for initial development. Anything may change at any time.
        > The public API should not be considered stable.

        ## [Unreleased]
    """  # noqa: E501, E800
    cbody = cdoc(body)
    return write_file('CHANGELOG.md', f'{cbody}\n')


def _update_gitignore(body: str) -> str:
    """List of things that should be in a gitignore file.

    Args:
        body: The current contents of the gitignore file.

    Returns:
        The contents of the new gitignore file.
    """
    buffer: List[str] = []
    if 'm/.m' not in body:
        buffer.append('m/.m')
    entries = '\n'.join(buffer)
    return f'{body}{entries}\n'


def update_gitignore() -> OneOf[Issue, int]:
    """Update the gitignore file.

    Adds the m/.m directory to the list.

    Returns:
        A `OneOf` containing 0 or an Issue if it was unable to update the file.
    """
    return one_of(lambda: [
        0
        for _ in eval_cmd('touch .gitignore')
        for body in read_file('.gitignore')
        for _ in write_file('.gitignore', _update_gitignore(body))
    ])


def init_repo() -> OneOf[Issue, int]:
    """Initialize a repository with the basic project configurations.

    Returns:
        A `OneOf` containing 0 if successful or an `Issue`.
    """
    obra_path = Path('m/m.json').resolve()
    if obra_path.exists():
        CiTool.warn('delete m/m.json to restart the init process.')
        return Good(0)
    return one_of(lambda: [
        0
        for _ in create_m_config()
        for _ in update_gitignore()
        for _ in create_changelog()
    ])



from dataclasses import dataclass
from distutils.version import StrictVersion

from ..core import Good, Issue, OneOf, issue, json, one_of
from ..core.io import JsonStr
from .types import (
    GitFlowConfig,
    MFlowConfig,
    Workflow,
    read_git_flow,
    read_m_flow,
    read_workflow,
)


def _handle_release_pr(ver_gt_latest: bool) -> str:
    return '' if ver_gt_latest else 'version needs to be bumped'


def _handle_release(ver_gt_latest: bool) -> str:
    return '' if ver_gt_latest else 'version was not bumped during release pr'


def _handle_non_release(ver_lt_latest: bool, ver_gt_latest: bool) -> str:
    if ver_lt_latest:
        return 'version is behind (Branch may need to be updated)'
    if ver_gt_latest:
        return 'version is ahead (Revert configuration change)'
    return ''


@dataclass
class Config(JsonStr):
    """Object to store the m project configuration."""

    # pylint: disable=too-many-instance-attributes
    owner: str
    repo: str
    version: str
    m_dir: str
    workflow: Workflow
    git_flow: GitFlowConfig
    m_flow: MFlowConfig
    build_tag_with_version: bool = False

    def uses_git_flow(self):
        """Check if configuration is using the git flow.

        Returns:
            True if workflow is the git flow.
        """
        return self.workflow == Workflow.git_flow

    def uses_m_flow(self):
        """Check if configuration is using the m flow.

        Returns:
            True if workflow is the m flow.
        """
        return self.workflow == Workflow.m_flow

    def uses_free_flow(self):
        """Check if configuration is using the free flow.

        Returns:
            True if workflow is the free flow.
        """
        return self.workflow == Workflow.free_flow

    def verify_version(
        self,
        gh_latest: str,
        is_release_pr: bool,
        is_release: bool,
    ) -> OneOf[Issue, int]:
        """Verify that the configuration version is valid.

        Args:
            gh_latest:
                The version stored in `Github`. Checks are skipped if
                this value is empty.
            is_release_pr:
                Set to `True` if the build is done during a release pr.
            is_release:
                Set to `True` if the build is done during a release.

        Returns:
            A `OneOf` containing 0 if all is well, otherwise an `Issue`.
        """
        if not gh_latest:
            return Good(0)
        err_data = {
            'config_version': self.version,
            'gh_latest': gh_latest,
            'is_release': is_release,
            'is_release_pr': is_release_pr,
        }
        try:
            p_ver = StrictVersion(self.version)
        except Exception as ex:
            return issue('error parsing version', cause=ex, context=err_data)
        try:
            p_latest = StrictVersion(gh_latest)
        except Exception as ex:  # noqa: WPS440
            return issue('error parsing latest', cause=ex, context=err_data)
        ver_gt_latest = p_ver > p_latest
        ver_lt_latest = p_ver < p_latest
        msg: str = ''
        if is_release_pr:
            msg = _handle_release_pr(ver_gt_latest)
        elif is_release:
            msg = _handle_release(ver_gt_latest)
        else:
            msg = _handle_non_release(ver_lt_latest, ver_gt_latest)
        return issue(msg, context=err_data) if msg else Good(0)


def read_config(m_dir: str) -> OneOf[Issue, Config]:
    """Read an m configuration file.

    Args:
        m_dir: Directory containing `m.json`.

    Returns:
        A `OneOf` containing the `m` configuration or an `Issue`.
    """
    return one_of(lambda: [
        Config(
            owner,
            repo,
            version,
            m_dir,
            workflow,
            git_flow,
            m_flow,
            build_tag_with_version=with_version,
        )
        for m_cfg in json.read_json(f'{m_dir}/m.json')
        for owner, repo in json.multi_get(m_cfg, 'owner', 'repo')
        for version in (m_cfg.get('version', '0.0.0'),)
        for workflow in read_workflow(m_cfg.get('workflow', 'free-flow'))
        for git_flow in read_git_flow(m_cfg.get('gitFlow', {}))
        for m_flow in read_m_flow(m_cfg.get('mFlow', {}))
        for with_version in (m_cfg.get('build_tag_with_version', False),)
    ]).flat_map_bad(lambda x: issue('read_config failure', cause=x))



import inspect

from ...utils import run_main


def add_parser(sub_parser, raw):
    desc = """
        Display the current commit sha.

            $ m git current_sha
            74075a3ea5c9252a0f2b9fd6b095567b3b9b4028
    """
    sub_parser.add_parser(
        'current_sha',
        help='display the current commit sha',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )


def run(_):
    # pylint: disable=import-outside-toplevel
    from .... import git
    return run_main(git.get_current_commit_sha, print)



import json as builtin_json
from http import client as httplib
from typing import Any, Mapping, Optional
from urllib.parse import urlparse

from . import Issue, issue, one_of
from .fp import Good, OneOf
from .json import parse_json

STATUS_OK = 200
STATUS_REDIRECT = 300


def _get_connection(protocol: str, hostname: str) -> httplib.HTTPConnection:
    if protocol == 'https':
        return httplib.HTTPSConnection(hostname)  # noqa: S309
    return httplib.HTTPConnection(hostname)


def fetch(
    url: str,
    headers: Mapping[str, str],
    method: str = 'GET',
    body: Optional[str] = None,
) -> OneOf[Issue, str]:
    """Send an http(s) request.

    Args:
        url:
            The url to request.
        headers:
            The headers for the request. By default it sets the `user-agent`
            to "m".
        method:
            The request method type. Defaults to `GET`.
        body:
            The body of the request.

    Returns:
        A `OneOf` containing the raw response from the server or an Issue.
    """
    parts = urlparse(url)
    protocol, hostname, path = [parts.scheme, parts.netloc, parts.path]
    path = f'{path}?{parts.query}' if parts.query else path
    fetch_headers = {'user-agent': 'm', **headers}
    if body:
        fetch_headers['content-length'] = str(len(body))
    connection = _get_connection(protocol, hostname)
    ctxt = {'url': f'{hostname}{path}'}
    # See the next link for explanation disabling WPS440:
    #  https://github.com/wemake-services/wemake-python-styleguide/issues/1416
    try:
        connection.request(method, path, body, fetch_headers)
    except Exception as ex:
        return issue(f'{protocol} request failure', cause=ex, context=ctxt)
    try:
        res = connection.getresponse()
    except Exception as ex:  # noqa: WPS440
        return issue(f'{protocol} response failure', cause=ex, context=ctxt)
    try:
        res_body = res.read()
    except Exception as ex:  # noqa: WPS440
        return issue(f'{protocol} read failure', cause=ex, context=ctxt)
    code = res.getcode()
    if STATUS_OK <= code < STATUS_REDIRECT:
        return Good(res_body)
    return issue(
        f'{protocol} request failure ({code})',
        context={
            'body': body,
            'code': code,
            'res_body': str(res_body),
            **ctxt,
        },
    )


def fetch_json(
    url: str,
    headers: Mapping[str, str],
    method: str = 'GET',
    body_json: Any = None,
) -> OneOf[Issue, Any]:
    """Especialized fetch to deal with json data.

    Args:
        url:
            The url to request.
        headers:
            Additional headers for the request. By default it will add
            proper accept and content-type headers for json requests.
        method:
            The request method type. Defaults to `GET`.
        body_json:
            The data to send to the server (python object).

    Returns:
        A `OneOf` containing a json parsed response from the server or an
        Issue.
    """
    body = builtin_json.dumps(body_json) if body_json else None
    fetch_headers = {
        'accept': 'application/json',
        'content-type': 'application/json',
        **headers,
    }
    return one_of(lambda: [
        response
        for payload in fetch(url, fetch_headers, method, body)
        for response in parse_json(payload)
    ])



import re
from typing import Any, List, Optional

from ..core import one_of
from ..core.fp import Good, OneOf
from ..core.issue import Issue
from ..core.json import get
from . import api
from .ci_dataclasses import (
    AssociatedPullRequest,
    Author,
    Commit,
    CommitInfo,
    GithubCiRunInfo,
    PullRequest,
    Release,
)
from .ci_graph_queries import LATEST_RELEASE, PULL_REQUEST, commit_query


def create_ci_query(
    pr_number: Optional[int] = None,
    include_commit: bool = False,
    include_release: bool = False,
) -> str:
    """Create github graphql query."""
    items: List[str] = []
    params: List[str] = ['$owner: String!', '$repo: String!']
    if include_commit:
        include_pr = pr_number is None
        items.append(commit_query(include_pr, True))
        params.append('$sha: String')
    if pr_number:
        items.append(PULL_REQUEST)
        params.append('$pr: Int!')
        params.append('$fc: Int!')
    if include_release:
        items.append(LATEST_RELEASE)
    items_str = '\n'.join(items)
    params_str = ', '.join(params)
    return f'''query ({params_str}) {{
      repository(owner:$owner, name:$repo) {{
        {items_str}
      }}
    }}'''


def _parse_commit_message(msg: str, sha: str) -> str:
    match = re.match('^Merge (.*) into (.*)$', msg)
    if match:
        return match.groups()[0]
    return sha


def get_build_sha(
    token: str,
    owner: str,
    repo: str,
    sha: str,
    get_sha: bool = True,
) -> OneOf[Issue, str]:
    """When building prs, we are not given the actual sha of the commit.

    Instead, we get the sha of the merge commit. This will give us the
    sha that we are looking for.
    """
    if not get_sha:
        return Good(sha)
    params = ['$owner: String!', '$repo: String!', '$sha: String!']
    params_str = ', '.join(params)
    query = f'''query ({params_str}) {{
      repository(owner:$owner, name:$repo) {{
        {commit_query(False, False)}
      }}
    }}'''
    variables = {
        'owner': owner,
        'repo': repo,
        'sha': sha,
    }
    return one_of(
        lambda: [
            _parse_commit_message(data, sha)
            for res in api.graphql(token, query, variables)
            for data in get(res, 'repository.commit.message')
        ],
    )


def get_raw_ci_run_info(
    token: str,
    commit_info: CommitInfo,
    pr_number: Optional[int],
    file_count: int,
    include_release: bool,
    get_sha: bool = True,
) -> OneOf[Issue, Any]:
    """Retrieve the information of the given Github PR."""
    query = create_ci_query(pr_number, True, include_release)
    owner, repo, sha = [commit_info.owner, commit_info.repo, commit_info.sha]
    variables = {'owner': owner, 'repo': repo, 'sha': sha, 'fc': file_count}
    if pr_number:
        variables['pr'] = pr_number
    return one_of(lambda: [
        data
        for variables['sha'] in get_build_sha(token, owner, repo, sha, get_sha)
        for res in api.graphql(token, query, variables)
        for data in get(res, 'repository')
    ])


def _get_release(raw: Any) -> OneOf[Issue, Optional[Release]]:
    releases = raw.get('releases')
    if releases:
        nodes = releases.get('nodes')
        if nodes:
            node = nodes[0]
            return Good(
                Release(
                    node['name'],
                    node['tagName'],
                    node['publishedAt'],
                ),
            )
    return Good(None)


def _get_commit(owner: str, repo: str, raw: Any) -> OneOf[Issue, Commit]:
    commit = raw['commit']
    sha = commit['oid']
    info = Commit(
        author_login=get(commit, 'author.user.login').get_or_else(''),
        short_sha=sha[:7],
        sha=sha,
        message=commit['message'],
        url=f'https://github.com/{owner}/{repo}/commit/{sha}',
    )
    nodes = get(commit, 'associatedPullRequests.nodes').get_or_else([])
    if nodes:
        pr = nodes[0]
        author = pr['author']
        info.associated_pull_request = AssociatedPullRequest(
            author=Author(
                login=author['login'],
                avatar_url=author['avatarUrl'],
                email=author['email'],
            ),
            merged=pr['merged'],
            pr_number=pr['number'],
            target_branch=pr['baseRefName'],
            target_sha=pr['baseRefOid'],
            pr_branch=pr['headRefName'],
            pr_sha=pr['headRefOid'],
            title=pr['title'],
            body=pr['body'],
        )
    return Good(info)


def _get_pull_request(
    raw: Any,
    pr_number: Optional[int],
) -> OneOf[Issue, Optional[PullRequest]]:
    pr = raw.get('pullRequest')
    if not pr:
        return Good(None)
    author = pr['author']
    return Good(
        PullRequest(
            author=Author(
                login=author['login'],
                avatar_url=author['avatarUrl'],
                email=author['email'],
            ),
            pr_number=pr_number or 0,
            pr_branch=pr['headRefName'],
            target_branch=pr['baseRefName'],
            target_sha=pr['baseRefOid'],
            url=pr['url'],
            title=pr['title'],
            body=pr['body'],
            file_count=pr['files']['totalCount'],
            files=[x['path'] for x in pr['files']['nodes']],
            is_draft=pr['isDraft'],
        ),
    )


def get_ci_run_info(
    token: str,
    commit_info: CommitInfo,
    pr_number: Optional[int],
    file_count: int,
    include_release: bool,
) -> OneOf[Issue, GithubCiRunInfo]:
    """Transform the result from get_raw_ci_run_info to a GithubCiRunInfo."""
    raw_res = get_raw_ci_run_info(
        token,
        commit_info,
        pr_number,
        file_count,
        include_release,
    )
    return one_of(
        lambda: [
            GithubCiRunInfo(commit, pr, release)
            for raw in raw_res
            for release in _get_release(raw)
            for commit in _get_commit(commit_info.owner, commit_info.repo, raw)
            for pr in _get_pull_request(raw, pr_number)
        ],
    )



import inspect


def add_parser(sub_parser, raw):
    desc = 'open a block to group several messages in the build log'
    parser = sub_parser.add_parser(
        'open',
        help='open block',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    parser.add_argument('name', type=str, help='block name to open')
    parser.add_argument('description', type=str, help='block description')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....core.io import CiTool
    CiTool.open_block(arg.name, arg.description)
    return 0



from typing import Any, Dict, List, cast

from ....core import Good, Issue, OneOf, json
from ..core.types import FileReport, Violation


def read_payload(payload: str) -> OneOf[Issue, List[FileReport]]:
    """Transform a pylint payload to a list of `FileReport` instances.

    Args:
        payload: The raw payload from pylint.

    Returns:
        A `OneOf` containing an `Issue` or a list of `FileReport` instances.
    """
    res = json.parse_json(payload)
    if res.is_bad:
        return res
    violations = cast(List[Any], res.value)
    report: Dict[str, List[Violation]] = {}
    for x in violations:
        violation = Violation(
            rule_id=x['symbol'],
            message=x['message'],
            line=int(x['line']),
            column=int(x['column']),
            file_path=x['path'],
        )
        if violation.file_path not in report:
            report[violation.file_path] = []
        report[violation.file_path].append(violation)
    return Good([
        FileReport(file_path=name, violations=violations)
        for name, violations in report.items()
    ])



from argparse import ArgumentParser
from inspect import cleandoc
from types import MappingProxyType
from typing import Any, List, Tuple

STORE_TRUE = MappingProxyType({'action': 'store_true'})

Arg = Tuple[List[str], str, Any]


def add_arguments(parser: ArgumentParser, args: Tuple[Arg, ...]) -> None:
    """Add arguments to the parser.

    Args:
        parser:
            An argparse ArgumentParser instance.
        args:
            A list of arguments. Each entry is composed of 3 items:
                - The names of the arguments ([-s, '--long'])
                - The help string (may be multilined)
                - A dictionary with extra options to `add_argument`.
    """
    for names, help_str, extra in args:
        parser.add_argument(*names, help=cleandoc(help_str), **extra)



import json
import math
import os
import sys
from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Optional, TextIO, Type, Union, cast

from .. import git
from . import issue, one_of
from .fp import Good, OneOf
from .issue import Issue


def format_seconds(number_of_seconds: Union[int, float]) -> str:
    """Return a string representing the number of seconds in a friendly
    format: Xd:Xh:Xm:Xs:Xms. """
    milliseconds = int(math.floor(number_of_seconds * 1000))
    _ms = milliseconds % 1000
    seconds = int(math.floor(milliseconds / 1000))
    sec = seconds % 60
    minutes = int(math.floor(seconds / 60))
    mins = minutes % 60
    hours = int(math.floor(minutes / 60))
    hrs = hours % 24
    days = int(math.floor(hours / 24))

    entries = [
        f'{days}d' if days else '',
        f'{hrs}h' if hours else '',
        f'{mins}m' if mins else '',
        f'{sec}s' if sec else '',
        f'{_ms}ms' if _ms else '',
    ]
    return ':'.join([x for x in entries if x]) or '0s'


def env(name: str, def_val: str = '') -> str:
    """Access an environment variable.

    Args:
        name: The name of the environment variable.
        def_val: The default value to return if not set. Defaults to empty.

    Returns:
        The value of the environment variable if set otherwise the `def_val`.
    """
    return os.environ.get(name, def_val)


def renv(key: str) -> OneOf[Issue, str]:
    """Require an environment variable to be defined.

    Args:
        key: The environment variable required to be defined.

    Returns:
        A `OneOf` with the value of the environment variable or an issue.
    """
    value = os.environ.get(key)
    # Value may still be an empty string, checking against None
    if value is not None:
        return Good(value)
    return issue(f'missing {key} in env')


def renv_vars(keys: List[str]) -> OneOf[Issue, List[str]]:
    """Require multiple env vars to be defined.

    Args:
        keys: The environment variables required to be defined.

    Returns:
        A `OneOf` with the values of the environment variables or an issue.
    """
    result: List[str] = []
    missing: List[str] = []
    for key in keys:
        value = os.environ.get(key)
        if value is None:
            missing.append(key)
        else:
            result.append(value)
    if missing:
        mstr = ', '.join(missing)
        return issue(f'missing [{mstr}] in env')
    return Good(result)


def read_file(filename: str) -> OneOf[Issue, str]:
    """Return a `Good` containing the contents of the file."""
    try:
        with open(filename, encoding='UTF-8') as fp:
            return Good(fp.read())
    except Exception as ex:
        return issue(
            'failed to read file',
            context={'filename': filename},
            cause=ex,
        )


def write_file(filename: str, contents: str) -> OneOf[Issue, int]:
    """Return a `Good` containing 0 if the file was written."""
    try:
        with open(filename, 'w', encoding='UTF-8') as fp:
            fp.write(contents)
        return Good(0)
    except Exception as ex:
        return issue(
            'failed to write file',
            context={'filename': filename},
            cause=ex,
        )


def _ver_str(major: int, minor: int, patch: int) -> str:
    return f'{major}.{minor}.{patch}'


def prompt_next_version(version: str, release_type: str) -> str:
    """Display the possible major, minor and patch versions and prompt the user
    to enter one of them. Return one of the versions.

    https://semver.org/
    """
    ver = version.split('-')[0]
    parts = [int(x) for x in ver.split('.')]
    patch = _ver_str(parts[0], parts[1], parts[2] + 1)
    if release_type == 'hotfix':
        return patch

    minor = _ver_str(parts[0], parts[1] + 1, 0)
    major = _ver_str(parts[0] + 1, 0, 0)
    valid = False
    options = [minor, major]
    msg = f'Current version is {version}. Enter one of the following:\n  '
    msg += '\n  '.join(options)
    result = ''
    while not valid:
        print(msg, file=sys.stderr)
        result = input('')
        if result in options:
            valid = True
    return result


def serialize(obj: Any) -> Any:
    """Return a serializable version of an object."""
    if isinstance(obj, Enum):
        return obj.value
    if hasattr(obj, '__dict__'):
        return obj.__dict__
    return f'[Non-Serializable {repr(obj)}]'


class JsonStr:
    """Base class to stringify dataclasses."""

    # pylint: disable=too-few-public-methods
    def __str__(self) -> str:
        return json.dumps(self.__dict__, default=serialize)


@dataclass
class EnvVars(JsonStr):
    """Class to store the values of the environment variables."""

    # pylint: disable=too-many-instance-attributes
    ci_env: bool = False
    github_token: str = ''
    server_url: str = ''
    run_id: str = ''
    run_number: str = ''
    run_url: str = ''
    git_branch: str = ''
    git_sha: str = ''
    triggered_by: str = ''
    triggered_by_email: str = ''
    triggered_by_user: str = ''


class CITool(ABC):
    """Class representing a continuous integration tool.

    We can use this class static methods to display local messages.
    """

    @staticmethod
    def env_vars() -> OneOf[Issue, EnvVars]:
        """Obtain basic environment variables."""
        res = EnvVars()
        return one_of(
            lambda: [
                res
                for res.git_branch in git.get_branch()
                for res.git_sha in git.get_current_commit_sha()
            ],
        )

    @staticmethod
    def open_block(name: str, description: str) -> None:
        """Define how a block is opened."""
        print(f'{name}: {description}')

    @staticmethod
    def close_block(_name: str) -> None:
        """Define how a block is closed."""

    @staticmethod
    def error(
        description: str,
        file: Optional[str] = None,
        line: Optional[str] = None,
        col: Optional[str] = None,
        stream: Optional[TextIO] = None,
    ) -> None:
        """Print an error message."""
        stream = stream or sys.stderr
        parts = [x for x in [file, line, col] if x]
        loc = ':'.join(parts)
        info = f'[{loc}]' if loc else ''
        print(f'error{info}: {description}', file=stream)

    @staticmethod
    def warn(
        description: str,
        file: Optional[str] = None,
        line: Optional[str] = None,
        col: Optional[str] = None,
        stream: Optional[TextIO] = None,
    ) -> None:
        """Print an warning message."""
        stream = stream or sys.stderr
        parts = [x for x in [file, line, col] if x]
        loc = ':'.join(parts)
        info = f'[{loc}]' if loc else ''
        print(f'warn{info}: {description}', file=stream)


class GithubActions(CITool):
    """Collection of methods used to communicate with Github."""

    @staticmethod
    def env_vars() -> OneOf[Issue, EnvVars]:
        """Read the environment variables from Github Actions."""
        res = EnvVars(
            ci_env=True,
            server_url='https://github.com',
        )
        return one_of(
            lambda: [
                res
                for [
                    repo,
                    res.run_id,
                    res.run_number,
                    res.github_token,
                    res.git_branch,
                    res.git_sha,
                    res.triggered_by,
                ] in renv_vars([
                    'GITHUB_REPOSITORY',
                    'GITHUB_RUN_ID',
                    'GITHUB_RUN_NUMBER',
                    'GITHUB_TOKEN',
                    'GITHUB_REF',
                    'GITHUB_SHA',
                    'GITHUB_ACTOR',
                ])
                for res.run_url in [
                    f'{res.server_url}/{repo}/actions/runs/{res.run_id}',
                ]
            ],
        ).flat_map_bad(lambda x: issue(
            'GH Actions env_vars failure', cause=cast(Issue, x),
        ))

    @staticmethod
    def open_block(name: str, _description: str) -> None:
        """Group log lines. There is no description field and it does not
        support nesting.

        https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#grouping-log-lines
        """
        print(f'::group::{name}')

    @staticmethod
    def close_block(_name: str) -> None:
        """Since Github Actions does not support nesting blocks, all this does
        is close the current block."""
        print('::endgroup::')

    @staticmethod
    def error(
        description: str,
        file: Optional[str] = None,
        line: Optional[str] = None,
        col: Optional[str] = None,
        stream: Optional[TextIO] = None,
    ) -> None:
        """Print an error message.

        https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#setting-an-error-message
        """
        stream = stream or sys.stderr
        loc = f'file={file},line={line},col={col}' if file else ''
        info = f' {loc}' if loc else ''
        print(f'::error{info}::{description}', file=stream)

    @staticmethod
    def warn(
        description: str,
        file: Optional[str] = None,
        line: Optional[str] = None,
        col: Optional[str] = None,
        stream: Optional[TextIO] = None,
    ) -> None:
        """Print a warning message.

        https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#setting-a-warning-message
        """  # noqa
        stream = stream or sys.stderr
        loc = f'file={file},line={line},col={col}' if file else ''
        info = f' {loc}' if loc else ''
        print(f'::warning{info}::{description}', file=stream)


class Teamcity(CITool):
    """Collection of methods used to communicate with Teamcity."""

    @staticmethod
    def env_vars() -> OneOf[Issue, EnvVars]:
        """Read basic environment variables from Teamcity."""
        # WIP: Need to map the to the object
        return Good(EnvVars(ci_env=True))

    @staticmethod
    def escape_msg(msg):
        """Escapes certain characters so that Teamcity may be able to print
        correctly. See:

        https://www.jetbrains.com/help/teamcity/build-script-interaction-with-teamcity.html#BuildScriptInteractionwithTeamCity-Escapedvalues
        """  # noqa
        return msg \
            .replace('|', '||') \
            .replace("'", "|'") \
            .replace('[', '|[') \
            .replace(']', '|]') \
            .replace('\n', '|n')

    @staticmethod
    def open_block(name: str, description: str) -> None:
        """Prints a command for teamcity to begin a new block where subsequent
        output will be bundled under. See:

        https://www.jetbrains.com/help/teamcity/build-script-interaction-with-teamcity.html#BuildScriptInteractionwithTeamCity-BlocksofServiceMessages
        """  # noqa
        desc = Teamcity.escape_msg(description)
        print(f"##teamcity[blockOpened name='{name}' description='{desc}']")

    @staticmethod
    def close_block(name: str) -> None:
        """Closes a previously defined block.

        See `open_block`.
        """
        print(f"##teamcity[blockClosed name='{name}']")

    @staticmethod
    def error(
        description: str,
        file: Optional[str] = None,
        line: Optional[str] = None,
        col: Optional[str] = None,
        stream: Optional[TextIO] = None,
    ) -> None:
        """Print a message to teamcity so that the build may abort."""
        stream = stream or sys.stderr
        parts = [x for x in [file, line, col] if x]
        loc = ':'.join(parts)
        info = f'[{loc}]: ' if loc else ''
        desc = Teamcity.escape_msg(f'{info}{description}' or '')
        print(
            f"##teamcity[buildProblem description='{desc}']",
            file=stream,
        )

    @staticmethod
    def warn(
        description: str,
        file: Optional[str] = None,
        line: Optional[str] = None,
        col: Optional[str] = None,
        stream: Optional[TextIO] = None,
    ) -> None:
        """Print a warning message to teamcity."""
        stream = stream or sys.stderr
        parts = [x for x in [file, line, col] if x]
        loc = ':'.join(parts)
        info = f'[{loc}]: ' if loc else ''
        desc = Teamcity.escape_msg(f'{info}{description}' or '')
        print(
            f"##teamcity[message status='WARNING' text='{desc}']",
            file=stream,
        )


def get_ci_tool() -> Type[CITool]:
    """Return the current CI Tool based on the environment variables."""
    if env('GITHUB_ACTIONS'):
        return GithubActions
    if env('TC') or env('TEAMCITY'):
        return Teamcity
    return CITool


CiTool = get_ci_tool()


def error_block(issue_: str, stream: Optional[TextIO] = None) -> None:
    """Print an issue within an error block."""
    CiTool.open_block('error', '')
    print(issue_, file=stream or sys.stderr)
    CiTool.close_block('error')


def warn_block(issue_: str, stream: Optional[TextIO] = None) -> None:
    """Print an issue within a warning block."""
    CiTool.open_block('warning', '')
    print(issue_, file=stream or sys.stderr)
    CiTool.close_block('warning')



import inspect
import sys
from typing import Tuple, cast

from ....core import one_of
from ...argparse import STORE_TRUE, Arg, add_arguments
from ...utils import display_issue
from ...validators import validate_json_payload, validate_payload

DESC = """
    process the output of a compiler or linter to determine if the ci
    process should stop.

    examples:

        ~$ m ci celt -t eslint -c @config.json < <(eslint [dir] -f json)

        ~$ eslint [...options] > tmp.json
        ~$ m ci celt -t eslint @tmp.json -c '{"allowedEslintRules":{"semi":1}}'

    Depending on the tool that is chosen the configuration should have an
    entry of the form "allowed[ToolName]Rules" or "ignored[ToolName]Rules".
    Only the first letter of the tool should be capitalized to conform to
    the camelcase style.

    The entry should define a map of rule ids to the number of allowed
    violations. In the case of `ignored[ToolName]Rules` we may define
    the rule id and assign an explanation as to why its being ignored.

    In the examples above we use `@config.json`. This means it
    will read the file `config.json`. You can use any file that you want.
    One option is to use eslintrc.json or create a brand new file called
    `allowed_errors.json`.

    Each tool will have different outputs. See below information for each
    tool.

    - eslint: expects json output.

        eslint -f json [dir]

    - pycodestyle: expects default output.

        pycodestyle --format=default [dir]

    - flake8: expects default output.

        flake8 [dir]

    - pylint: expects json output

        pylint -f json --rcfile=[file] [dir]
"""

ARGS: Tuple[Arg, ...] = (
    (
        ['payload'],
        'data: @- (stdin), @filename (file), string. Defaults to @-',
        {'type': validate_payload, 'nargs': '?', 'default': '@-'},
    ),
    (
        ['-t', '--tool'],
        'name of a supported compiler/linter',
        {'required': True},
    ),
    (
        ['-c', '--config'],
        'config data: @filename (file), string',
        {'type': validate_json_payload, 'default': '{}'},  # noqa: P103
    ),
    (
        ['-m', '--max-lines'],
        'max number of error lines to print (Defaults to 5, -1 for all)',
        {'type': int, 'default': 5},
    ),
    (
        ['-r', '--file-regex'],
        'regex expression to filter files',
        {},
    ),
    (
        ['-p', '--file-prefix'],
        "replace file prefix with 'old1|old2:new'",
        {},
    ),
    (
        ['-i', '--ignore-error-allowance'],
        'set every error allowence to 0',
        {**STORE_TRUE},
    ),
    (
        ['-s', '--stats-only'],
        'display a dictionary with current total violations',
        {**STORE_TRUE},
    ),
    (
        ['-f', '--full-message'],
        'display the full error message',
        {**STORE_TRUE},
    ),
    (
        ['--traceback'],
        'display the exception traceback if available',
        {**STORE_TRUE},
    ),
)


def add_parser(sub_parser, raw):
    parser = sub_parser.add_parser(
        'celt',
        help='process the output of compiler/linter',
        formatter_class=raw,
        description=inspect.cleandoc(DESC),
    )
    add_arguments(parser, ARGS)


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....ci.celt.core.process import PostProcessor
    from ....ci.celt.core.types import Configuration, ProjectStatus
    from ....ci.celt.post_processor import get_post_processor
    from ....core.io import CiTool
    from ....core.issue import Issue

    config = Configuration(
        arg.max_lines,
        arg.full_message,
        arg.ignore_error_allowance,
        arg.file_regex,
        arg.file_prefix,
    )
    tool_either = get_post_processor(arg.tool, config)
    either = one_of(lambda: [
        project
        for tool in tool_either
        for project in tool.run(arg.payload, arg.config)
    ])
    if either.is_bad:
        Issue.show_traceback = arg.traceback
        display_issue(cast(Issue, either.value))
        return 1
    tool = cast(PostProcessor, tool_either.value)
    project = cast(ProjectStatus, either.value)
    output = (
        tool.stats_json(project)
        if arg.stats_only
        else tool.to_str(project)
    )
    print(output, file=sys.stderr)  # noqa: WPS421
    if project.error_msg:
        CiTool.error(project.error_msg)
    return project.status.value



from dataclasses import dataclass
from enum import Enum
from typing import Any, Mapping

from ..core import Good, Issue, OneOf, issue
from ..core.io import JsonStr


class Branches(Enum):
    """Default branches to use for the supported workflows."""

    master = 'master'
    develop = 'develop'
    release = 'release'
    hotfix = 'hotfix'

    def __str__(self) -> str:
        """Stringify the `Branches` enum.

        Returns:
            The value of the enum.
        """
        return str(self.value)


class Workflow(Enum):
    """Supported workflows."""

    git_flow = 'git_flow'
    m_flow = 'm_flow'
    free_flow = 'free_flow'

    def __str__(self) -> str:
        """Stringify the `Workflow` enum.

        Returns:
            The value of the enum.
        """
        return str(self.value)


@dataclass
class GitFlowConfig(JsonStr):
    """An object mapping branches for the git_flow workflow."""

    master_branch: str = str(Branches.master)
    develop_branch: str = str(Branches.develop)
    release_prefix: str = str(Branches.release)
    hotfix_prefix: str = str(Branches.hotfix)


@dataclass
class MFlowConfig(JsonStr):
    """An object mapping branches for the m_flow workflow."""

    master_branch: str = str(Branches.master)
    release_prefix: str = str(Branches.release)
    hotfix_prefix: str = str(Branches.hotfix)


def read_workflow(raw_name: str) -> OneOf[Issue, Workflow]:
    """Parse the workflow field.

    Args:
        raw_name:
            The name of the workflow in the configuration. Dashes are allowed.
            For instance `git-flow`, 'm-flow`, but an underscore will work too.

    Returns:
        A `OneOf` containing a `Workflow` enum or and `Issue`.
    """
    name = raw_name.lower().replace('-', '_')
    try:
        return Good(Workflow[name])
    except Exception as ex:
        return issue('invalid workflow', cause=ex)


def read_git_flow(prop_val: Mapping[str, Any]) -> OneOf[Issue, GitFlowConfig]:
    """Parse the gitFlow field.

    Args:
        prop_val: The dictionary value in the gitFlow field.

    Returns:
        A `OneOf` containing a GitFlowConfig instance.
    """
    config = GitFlowConfig(
        master_branch=prop_val.get('masterBranch', str(Branches.master)),
        develop_branch=prop_val.get('developBranch', str(Branches.develop)),
        release_prefix=prop_val.get('releasePrefix', str(Branches.release)),
        hotfix_prefix=prop_val.get('hotfixPrefix', str(Branches.hotfix)),
    )
    return Good(config)


def read_m_flow(prop_val: Mapping[str, Any]) -> OneOf[Issue, MFlowConfig]:
    """Parse the mFlow field.

    Args:
        prop_val: The dictionary value in the mFlow field.

    Returns:
        A `OneOf` containing a MFlowConfig instance.
    """
    config = MFlowConfig(
        master_branch=prop_val.get('masterBranch', str(Branches.master)),
        release_prefix=prop_val.get('releasePrefix', str(Branches.release)),
        hotfix_prefix=prop_val.get('hotfixPrefix', str(Branches.hotfix)),
    )
    return Good(config)



import inspect

from ...utils import env, run_main


def add_parser(sub_parser, raw):
    desc = """
        Create a commit status

        - https://docs.github.com/en/rest/reference/repos#create-a-commit-status

        example:

            $ m github status \\
                --owner jmlopez-rod \\
                --repo pysync \\
                --sha [sha] \\
                --context github-check \\
                --state pending \\
                --decription 'running checks'
    """  # noqa
    parser = sub_parser.add_parser(
        'status',
        help='create a commit status',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--owner',
        type=str,
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner (default: env.GITHUB_REPOSITORY_OWNER)',
    )
    add(
        '--repo',
        type=str,
        required=True,
        help='repo name',
    )
    add(
        '--sha',
        required=True,
        help='The commit sha',
    )
    add(
        '--context',
        required=True,
        help='unique identifier for the status (a name?)',
    )
    add(
        '--state',
        required=True,
        choices=['error', 'failure', 'pending', 'success'],
        help='the state of the status',
    )
    add(
        '--description',
        required=True,
        help='a short description of the status',
    )
    add('--url', help='URL to associate with this status')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....github.api import GithubShaStatus, commit_status
    return run_main(
        lambda: commit_status(
            arg.token,
            arg.owner,
            arg.repo,
            GithubShaStatus(
                sha=arg.sha,
                context=arg.context,
                state=arg.state,
                description=arg.description,
                url=arg.url,
            ),
        ),
    )



import inspect


def add_parser(sub_parser, raw):
    desc = """
        report an error

        example:

            ~$ m message error 'this is an error'
            ::error::this is an error
            ~$ echo $?
            1

        The procedure exits with non-zero code.
    """
    parser = sub_parser.add_parser(
        'error',
        help='report an error',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add('message', type=str, help='error message')
    add('-f', '--file', type=str, help='filename where error occurred')
    add('-l', '--line', type=str, help='line where error occurred')
    add('-c', '--col', type=str, help='column where error occurred')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....core.io import CiTool
    CiTool.error(arg.message, arg.file, arg.line, arg.col)
    return 1



import inspect

from ...utils import run_main


def add_parser(sub_parser, raw):
    desc = """
        Display a single word representing the current git status.

            $ m git status
            clean

        Statuses:

            unknown
            untracked
            stash
            clean
            ahead
            behind
            staged
            dirty
            diverged
            ?
    """
    sub_parser.add_parser(
        'status',
        help='display the current git status',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )


def run(_):
    # pylint: disable=import-outside-toplevel
    from .... import git
    return run_main(git.get_status, print)



from typing import Callable, Generic, Iterator, TypeVar, Union, cast

A = TypeVar('A')  # pylint: disable=invalid-name
B = TypeVar('B')  # pylint: disable=invalid-name
G = TypeVar('G')  # pylint: disable=invalid-name
LazyArg = Union[A, Callable[[], A]]


def lazy_arg(z_arg: LazyArg[A]) -> A:
    """Return the output of calling `z_arg` if it is function.

    Otherwise param is returned.

    Args:
        z_arg: A function or a value

    Returns:
        The value.
    """
    return z_arg() if callable(z_arg) else z_arg


class StopBadIteration(Exception):  # noqa: N818 - This is for internal use
    """Store a `Bad` instance."""

    def __init__(self, bad):
        """Initialize the Exception.

        Args:
            bad: The `Bad` value stored in a `OneOf`.
        """
        super().__init__()
        self.bad = bad


class OneOf(Generic[B, G]):
    """An instance of `OneOf` is an instance of either `Bad` or `Good`."""

    def __init__(self, bad: bool, one_of_value: Union[B, G]):
        """Initialize a `OneOf`.

        Args:
            bad:
                Set to `True` if the instance holds a `Bad` value.
            one_of_value:
                The value of the instance.
        """
        self.is_bad = bad
        self.value = one_of_value  # noqa: WPS110

    def __iter__(self) -> Iterator[G]:
        if self.is_bad:
            raise StopBadIteration(self)
        yield cast(G, self.value)

    def iter(self):
        """Shortcut to transform to a list: list(x.iter()).

        It will either contain a value or be an empty list.
        """
        if not self.is_bad:
            yield self.value

    def map(self, fct):
        """Apply the function to its value if this is a `Good` instance.

        Args:
            fct: The function to apply to the "Good" value.

        Returns:
            Itself if its a `Bad` otherwise another instance of `Good`.
        """
        return self if self.is_bad else Good(fct(self.value))

    def flat_map_bad(self, fct: Callable[[B], 'OneOf']) -> 'OneOf':
        """Apply the input function if this is a `Bad` value.

        Args:
            fct: The function to apply to the "Bad" value.

        Returns:
            Itself if its a `Good` otherwise another instance of `Bad`.
        """
        return fct(cast(B, self.value)) if self.is_bad else self

    def get_or_else(self, or_: LazyArg[G]) -> G:
        """Returns the value if its Good or the given argument if its a Bad."""
        return lazy_arg(or_) if self.is_bad else cast(G, self.value)


class Bad(OneOf[B, G]):
    """The bad side of the disjoint union."""

    def __init__(self, one_of_value):
        """Initialize a `Bad` instance.

        Args:
            one_of_value: The "bad" value.
        """
        super().__init__(bad=True, one_of_value=one_of_value)


class Good(OneOf[B, G]):
    """The good side of the disjoint union."""

    def __init__(self, one_of_value):
        """Initialize a `Good` instance.

        Args:
            one_of_value: The "good" value.
        """
        super().__init__(bad=False, one_of_value=one_of_value)



# noqa: WPS412
from .fp import Bad, Good, OneOf
from .issue import Issue
from .one_of import issue, one_of

__all__ = [  # noqa: WPS410
    'Issue',
    'issue',
    'one_of',
    'Good',
    'OneOf',
    'Bad',
]



from .cli.utils import run_cli
from .version import VERSION


def main_args(argp):
    argp.add_argument('--version', action='version', version=VERSION)


def main():
    run_cli(__file__, main_args)


if __name__ == '__main__':  # pragma: no cover
    main()



import inspect
import re


def add_parser(sub_parser, raw):
    desc = """
        display an npm tag assigned to given version

        examples:

            ~$ m ci npm_tag 0.0.0-develop.b123
            develop

            ~$ m ci npm_tag 0.0.0-pr1234.b123
            pr1234

            ~$ m ci npm_tag 2.0.1-rc1234.b123
            next

            ~$ m ci npm_tag 2.0.1
            latest
    """
    parser = sub_parser.add_parser(
        'npm_tag',
        help='display an npm tag',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    parser.add_argument('version', type=str, help='npm package version')


def run(arg):
    regex = r'\d.\d.\d-(.*)\.(.*)'
    matches = re.match(regex, arg.version)
    if matches:
        tag, _ = matches.groups()
        if tag.startswith('rc') or tag.startswith('hotfix'):
            print('next')
        else:
            print(tag)
    else:
        print('latest')



import inspect

from ...utils import env, run_main
from ...validators import validate_payload


def add_parser(sub_parser, raw):
    desc = """
        Create a pull request

        https://docs.github.com/en/rest/reference/pulls#create-a-pull-request

        example:

            $ m github create_pr \\
                --owner jmlopez-rod \\
                --repo repo \\
                --head feature_branch \\
                --base master \\
                --title 'PR Title' \\
                @file_with_pr_body | m json
    """
    parser = sub_parser.add_parser(
        'create_pr',
        help='create a pull request',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--owner',
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner (default: env.GITHUB_REPOSITORY_OWNER)',
    )
    add(
        '--repo',
        required=True,
        help='repo name',
    )
    add(
        '--head',
        required=True,
        help='name of the branch where the changes are implemented.',
    )
    add(
        '--base',
        required=True,
        help='name of the branch you want the changes pulled into',
    )
    add(
        '--title',
        required=True,
        help='pull request title',
    )
    add(
        'body',
        type=validate_payload,
        nargs='?',
        default='@-',
        help='data: @- (stdin), @filename (file), string. Defaults to @-',
    )


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....github.api import GithubPullRequest, create_pr
    return run_main(lambda: create_pr(
        arg.token,
        arg.owner,
        arg.repo,
        GithubPullRequest(arg.title, arg.body, arg.head, arg.base),
    ))



from ...utils import run_main


def add_parser(sub_parser, raw):
    sub_parser.add_parser(
        'init',
        help='initialize an m project',
        formatter_class=raw,
        description='Create the necessary files for an m project.',
    )


def run(_arg):
    # pylint: disable=import-outside-toplevel
    from ....ci.init import init_repo
    return run_main(init_repo)



import inspect

from ...utils import env, run_main


def add_parser(sub_parser, raw):
    desc = """
        Retrieve a pull request information.

        example:

            $ m github pr --owner microsoft --repo typescript 44710 | m json
            {
              "headRefName": "ReduceExceptions",
              "headRefOid": "d9ae52cf49732a2d45b6cb7f4069205c88af39eb",
              "baseRefName": "main",
              "baseRefOid": "6452cfbad0afcc6d09b75e0a1e32da1d07e0b7ca",
              "title": "Reduce exceptions",
              "body": "...
    """
    parser = sub_parser.add_parser(
        'pr',
        help='get information on a pull request',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--owner',
        type=str,
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner (default: env.GITHUB_REPOSITORY_OWNER)',
    )
    add(
        '--repo',
        type=str,
        required=True,
        help='repo name',
    )
    add(
        '--files',
        type=int,
        default=10,
        help='max number of files to retrieve (default: 10)',
    )
    add('pr_number', type=int, help='pull request number')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....github.cli import get_pr_info
    return run_main(lambda: get_pr_info(
        arg.token,
        arg.owner,
        arg.repo,
        arg.pr_number,
        arg.files,
    ))



from inspect import cleandoc as cdoc
from typing import Tuple

from ..argparse import STORE_TRUE, Arg, add_arguments
from ..validators import validate_json_payload

DESC = """
    Format a json payload.

        $ echo '{"a":99}' | m json
        {
            "a": 99
        }

    similar to `python -m json.tool` but instead it uses 2 spaces
    for indentation.
"""
ARGS: Tuple[Arg, ...] = (
    (
        ['payload'],
        'json data: @- (stdin), @filename (file), string. Defaults to @-',
        {'type': validate_json_payload, 'nargs': '?', 'default': '@-'},
    ),
    (
        ['--sort-keys'],
        'sort the output of dictionaries alphabetically by key',
        {**STORE_TRUE},
    ),
)


def add_parser(sub_parser, raw):
    parser = sub_parser.add_parser(
        'json',
        help='format json data',
        formatter_class=raw,
        description=cdoc(DESC),
    )
    add_arguments(parser, ARGS)


def run(arg):
    # pylint: disable=import-outside-toplevel
    import json
    import sys
    json.dump(arg.payload, sys.stdout, indent=2, sort_keys=arg.sort_keys)
    sys.stdout.write('\n')
    return 0



import inspect

from ...utils import env, run_main


def add_parser(sub_parser, raw):
    desc = """
        Retrieve the latest release.

        example:

            $ m github latest_release --owner microsoft --repo typescript

    """
    parser = sub_parser.add_parser(
        'latest_release',
        help='get the latest release for a repo',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--owner',
        type=str,
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner (default: env.GITHUB_REPOSITORY_OWNER)',
    )
    add(
        '--repo',
        required=True,
        help='repo name',
    )


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....github.cli import get_latest_release
    return run_main(
        lambda: get_latest_release(
            arg.token,
            arg.owner,
            arg.repo,
        ), print,
    )



import inspect


def add_parser(sub_parser, raw):
    desc = """
        report an warning

        example:

            ~$ m message warn 'this is a warning' -f app.js -l 1 -c 5
            ::warning file=app.js,line=1,col=5::Missing semicolon
    """
    parser = sub_parser.add_parser(
        'warn',
        help='report a warning',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add('message', type=str, help='warning message')
    add('-f', '--file', type=str, help='filename where warning occurred')
    add('-l', '--line', type=str, help='line where warning occurred')
    add('-c', '--col', type=str, help='column where warning occurred')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....core.io import CiTool
    CiTool.warn(arg.message, arg.file, arg.line, arg.col)
    return 0



import shlex
from subprocess import STDOUT, CalledProcessError, check_output  # noqa: S404

from . import Issue, issue
from .fp import Good, OneOf


def eval_cmd(cmd: str) -> OneOf[Issue, str]:
    """Evaluate a bash command and return its output.

    Args:
        cmd: The shell command to evaluate.

    Returns:
        The output of the command (or an Issue if the command failed).
    """
    command = shlex.split(cmd)
    try:
        out = check_output(command, stderr=STDOUT, shell=False).decode()  # noqa: S603,E501
    except CalledProcessError as ex:
        out = ex.output.decode()
        return issue(
            'command returned a non zero exit code',
            context={'cmd': cmd, 'output': out},
        )
    return Good(out.strip())



import inspect

from ...utils import run_main


def add_parser(sub_parser, raw):
    desc = """
        Display the very first commit sha in the repository.

            $ m git first_sha
            bf286e270e13c75dfed289a3921289092477c058
    """
    sub_parser.add_parser(
        'first_sha',
        help='display the first commit sha',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )


def run(_):
    # pylint: disable=import-outside-toplevel
    from .... import git
    return run_main(git.get_first_commit_sha, print)



import inspect

from ...utils import env, run_main


def add_parser(sub_parser, raw):
    desc = """
        Provide the build sha. That is, provide GITHUB_SHA (a merge commit)
        and we will get the sha of the actual commit we thought we were
        building.

        example:

    """  # noqa
    parser = sub_parser.add_parser(
        'build_sha',
        help='get the correct commit sha',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--owner',
        type=str,
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner (default: env.GITHUB_REPOSITORY_OWNER)',
    )
    add(
        '--repo',
        type=str,
        required=True,
        help='repo name',
    )
    add(
        '--sha',
        type=str,
        required=True,
        help='commit sha',
    )


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....github.ci import get_build_sha
    return run_main(
        lambda: get_build_sha(arg.token, arg.owner, arg.repo, arg.sha),
        handle_result=print,
    )



import inspect

from ...utils import env, run_main


def add_parser(sub_parser, raw):
    desc = """
        Create a release in Github.

        - https://docs.github.com/en/rest/reference/repos#create-a-release

        example:

            $ m github release \\
                --owner jmlopez-rod \\
                --repo pysync \\
                --version 1.0.0
    """
    parser = sub_parser.add_parser(
        'release',
        help='create a github release',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--owner',
        type=str,
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner (default: env.GITHUB_REPOSITORY_OWNER)',
    )
    add(
        '--repo',
        type=str,
        required=True,
        help='repo name',
    )
    add(
        '--version',
        type=str,
        required=True,
        help='version to release',
    )
    add(
        '--branch',
        type=str,
        default=None,
        help='The branch where the git tag will be created',
    )


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....github.api import create_release
    return run_main(
        lambda: create_release(
            arg.token,
            arg.owner,
            arg.repo,
            arg.version,
            arg.branch,
        ),
    )



import json
import sys
from collections.abc import Mapping
from contextlib import suppress
from typing import Any, List
from typing import Mapping as Map
from typing import Optional, Union, cast

from .fp import Good, OneOf
from .io import CITool
from .issue import Issue
from .one_of import issue


def read_json(
    filename: Optional[str],
    error_if_empty: bool = False,
) -> OneOf[Issue, Any]:
    """Return a `Good` containing the parsed contents of the json file."""
    try:
        empty: str = '' if error_if_empty else 'null'
        if filename is None:
            return Good(json.loads(sys.stdin.read() or empty))
        with open(filename, encoding='UTF-8') as file_handle:
            return Good(json.loads(file_handle.read() or empty))
    except Exception as ex:
        return issue(
            'failed to read json file',
            context={'filename': filename or 'SYS.STDIN'},
            cause=ex,
        )


def parse_json(
    data: str,
    error_if_empty: bool = False,
) -> OneOf[Issue, Any]:
    """Return a `Good` containing the parsed contents of the json string."""
    try:
        empty = '' if error_if_empty else 'null'
        return Good(json.loads(data or empty))
    except Exception as ex:
        return issue('failed to parse the json data', cause=ex)


def get(obj: Any, key_str: str) -> OneOf[Issue, Any]:
    """Return the value based on the `key_str` specified. The following
    are equivalent:

    ```
    obj['a']['b']['c'] <=> get(obj, 'a.b.c').value
    ```

    provided that the keys `a`, `b` and `c` are valid. Returns a `Good` if
    the value we want is available, otherwise it returns a `Bad` with the
    path that returned `None`.
    """
    keys = key_str.split('.')
    current = obj
    for num, key in enumerate(keys):
        new_key: Union[str, int] = key
        with suppress(ValueError):
            new_key = int(key)
        try:
            current = current[new_key]
        except KeyError:
            pth = '.'.join(keys[:num + 1])
            return issue(f'`{pth}` path was not found')
        except Exception as ex:
            pth = '.'.join(keys[:num])
            if not isinstance(current, Mapping):
                context = pth or current
                return issue(f'`{context}` is not a dict')
            return issue(f'{pth} resulted in an error', cause=ex)
    return Good(current)


def multi_get(
    obj: object,
    *keys: str,
) -> OneOf[Issue, List[Any]]:
    """Call `get` for every input specified by `keys`. It collects the invalid
    keys and returns an `Issue`.

    multi_get(obj, 'a', 'a.b', 'a.b.c')
    """
    result = []
    failures = []
    for key in keys:
        res = get(obj, key)
        if res.is_bad:
            failures.append(
                Issue(
                    message=f'key lookup failure: `{key}`',
                    cause=res.value,
                    include_traceback=False,
                ),
            )
        else:
            result.append(res.value)
    if failures:
        return issue(
            'multi_get key retrieval failure',
            context=[x.to_dict() for x in failures],
            include_traceback=False,
        )
    return Good(result)


def _to_str(obj):
    if isinstance(obj, bool):
        return f'{obj}'.lower()
    if not obj:
        return 'null'
    return f'{obj}'


def jsonq(
    obj: Map[str, Any],
    separator: str,
    display_warning: bool,
    *key_str: str,
) -> int:
    """Print the values obtained from `multi_get` to stdout.

    Returns 0 if all the keys are available. Returns non-zero if there
    are problems.
    """
    result = multi_get(obj, *key_str)
    if result.is_bad:
        val = cast(Issue, result.value)
        if display_warning:
            CITool.warn(val.message)
        else:
            CITool.error(val.message)
        print(result.value, file=sys.stderr)
        return 1
    values = [_to_str(obj) for obj in cast(List[Any], result.value)]
    print(separator.join(values))
    return 0



def add_parser(sub_parser, raw):
    parser = sub_parser.add_parser(
        'bump_version',
        help='prompt for the next version',
        formatter_class=raw,
        description='Prompt user for the next valid semantic version',
    )
    add = parser.add_argument
    add(
        '--type',
        required=True,
        choices=['release', 'hotfix'],
        help='verification type',
    )
    add('version', help='version to bump')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....core.io import prompt_next_version
    next_ver = prompt_next_version(arg.version, arg.type)
    print(next_ver)  # noqa: WPS421
    return 0



import inspect

from ...utils import run_main


def add_parser(sub_parser, raw):
    desc = """
        Used during a release setup or hotfix setup. We want to make sure
        that we are working on the correct branch depending on verification
        we want to make and the workflow that we are using.
    """
    parser = sub_parser.add_parser(
        'assert_branch',
        help='assert that we are working on the correct branch',
        formatter_class=raw,
        description=inspect.cleandoc(desc),
    )
    add = parser.add_argument
    add(
        '--type',
        required=True,
        choices=['release', 'hotfix'],
        help='verification type',
    )
    add('m_dir', type=str, help='m project directory')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....ci.assert_branch import assert_branch
    return run_main(lambda: assert_branch(arg.type, arg.m_dir))



from ...utils import run_main


def add_parser(sub_parser, raw):
    parser = sub_parser.add_parser(
        'release_setup',
        help='modify the config and changelog files',
        formatter_class=raw,
        description='Update the config and changelog files',
    )
    add = parser.add_argument
    add(
        '--changelog',
        type=str,
        default='CHANGELOG.md',
        help='CHANGELOG filename',
    )
    add('m_dir', type=str, help='m project directory')
    add('new_ver', type=str, help='the new version')


def run(arg):
    # pylint: disable=import-outside-toplevel
    from ....ci.release_setup import release_setup
    return run_main(
        lambda: release_setup(
            arg.m_dir,
            arg.new_ver,
            arg.changelog,
        ),
    )



from typing import Dict, List, Set, cast

from ..core import Issue, issue, one_of
from ..core.fp import Good, OneOf
from .cli import get_dist_tags, remove_dist_tag


def clean_tags(pkg: str) -> OneOf[Issue, List[str]]:
    """Remove tags from a package that point to empty versions.

    Args:
        pkg: The name of the npm package.

    Returns:
        A `OneOf` containing a summary of removed tags or an Issue.
    """
    return one_of(lambda: [
        summary
        for tag_map in get_dist_tags(pkg)
        for summary in remove_tags(pkg, find_empty_tags(tag_map))
    ])


def find_empty_tags(tag_map: Dict[str, str]) -> Set[str]:
    """Create a set containing the npm tags that are empty.

    Args:
        tag_map: A dictionary mapping npm tags to versions.

    Returns:
        A set of npm tags that map to empty strings.
    """
    return {
        tag_name
        for tag_name, version in tag_map.items()
        if not version
    }


def remove_tags(pkg: str, tags: Set[str]) -> OneOf[Issue, List[str]]:
    """Call `npm dist-tag` to remove npm tags.

    See https://docs.npmjs.com/cli/v8/commands/npm-dist-tag

    Args:
        pkg: The npm package.
        tags: A set containing the tags to remove.

    Returns:
        A `OneOf` containing a list of the output for each removed tag or an
        issue with a summary.
    """
    issues: List[Issue] = []
    removed: List[str] = []
    for tag in tags:
        cmd_result = remove_dist_tag(pkg, tag)
        if cmd_result.is_bad:
            issues.append(cast(Issue, cmd_result.value))
        else:
            removed.append(cast(str, cmd_result.value))
    if issues:
        return issue('dist-tag rm issues', context={
            'issues': issues,
            'removed': removed,
        })
    return Good(removed)



from inspect import cleandoc

_DESC = """
    CI/CD tools provide ways of interacting with them via shell messages.

    Github Actions:
        https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions

    Teamcity:
        https://www.jetbrains.com/help/teamcity/service-messages.html

    The commands provided here attempt to create an interface to
    display those messages.
"""

meta = {
    'help': 'display messages',
    'description': cleandoc(_DESC),
}
